""" Parser file for Thermo Fisher Qubit Flex Adapter"""
from __future__ import annotations

from typing import Any, TypeVar

import numpy as np
import pandas as pd

from allotropy.allotrope.converter import add_custom_information_document
from allotropy.allotrope.models.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    ContainerType,
    DataSystemDocument,
    DeviceSystemDocument,
    FluorescenceEmissionSpectrumDetectionMeasurementDocumentItems,
    FluorescencePointDetectionDeviceControlAggregateDocument,
    FluorescencePointDetectionDeviceControlDocumentItem,
    FluorescencePointDetectionMeasurementDocumentItems,
    LuminescencePointDetectionMeasurementDocumentItems,
    MeasurementAggregateDocument,
    Model,
    SampleDocument,
    SpectrophotometryAggregateDocument,
    SpectrophotometryDocumentItem,
    UltravioletAbsorbancePointDetectionMeasurementDocumentItems,
    UltravioletAbsorbanceSpectrumDetectionMeasurementDocumentItems,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueMicrogramPerMicroliter,
    TQuantityValueMicrogramPerMilliliter,
    TQuantityValueMicroliter,
    TQuantityValueMilligramPerMilliliter,
    TQuantityValueNanogramPerMicroliter,
    TQuantityValueNanogramPerMilliliter,
    TQuantityValuePicogramPerMilliliter,
    TQuantityValueRelativeFluorescenceUnit,
    TQuantityValueUnitless,
)
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.thermo_fisher_qubit_flex import constants
from allotropy.parsers.thermo_fisher_qubit_flex.thermo_fisher_qubit_flex_reader import (
    ThermoFisherQubitFlexReader,
)
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.vendor_parser import VendorParser

DataType = TypeVar("DataType")

CONCENTRATION_UNIT_TO_TQUANTITY = {
    "ug/uL": TQuantityValueMicrogramPerMicroliter,
    "ug/mL": TQuantityValueMicrogramPerMilliliter,
    "mg/mL": TQuantityValueMilligramPerMilliliter,
    "ng/uL": TQuantityValueNanogramPerMicroliter,
    "ng/mL": TQuantityValueNanogramPerMilliliter,
    "pg/uL": TQuantityValuePicogramPerMilliliter,
}


def _get_concentration_value(
    data_frame: pd.DataFrame, column: str, units_column: str, row: int
) -> DataType | None:
    """
    Retrieves the value and its unit from the specified columns and row in the DataFrame. If units are not there, replace it with unitless" unit.

    parameters:
    data_frame (pd.DataFrame): The DataFrame from which to retrieve the value.
    column (str): The column name from which to retrieve the value.
    units_column (str): The column name from which to retrieve the unit.
    row (int): The row index from which to retrieve the value.

    Returns:
    Optional[DataType|None]: The concentration value converted to the appropriate data type, or None if the units are not available or invalid.
    """
    units = _get_value(data_frame, units_column, row)
    if units is None:
        units = ""
    datatype = CONCENTRATION_UNIT_TO_TQUANTITY.get(units, TQuantityValueUnitless)
    return _get_property_value(data_frame, column, row, datatype)


def _get_value(data_frame: pd.DataFrame, column: str, row: int) -> Any | None:
    """
    Retrieves the value from a specified column and row in a DataFrame, handling NaNs
    and converting certain numpy types to native Python types.

    Parameters:
    data_frame (pd.DataFrame): The DataFrame from which to retrieve the value.
    column (str): The column name from which to retrieve the value.
    row (int): The row index from which to retrieve the value.

    Returns:
    Optional[Any|None]: The value from the specified cell converted to the appropriate Python type.
                   Returns None if the column does not exist or the value is NaN.
    """
    if column not in data_frame.columns:
        return None
    value = data_frame[column][row]

    if pd.isna(value):
        return None
    if isinstance(value, np.int64):
        return int(value)
    if isinstance(value, np.float64):
        return float(value)
    return value


def _get_property_value(
    data_frame: pd.DataFrame, column: str, row: int, datatype: type
) -> DataType | None:
    """
    Retrieves the value from a specified column and row in a DataFrame and converts it
    to the specified datatype.

    Parameters:
    data_frame (pd.DataFrame): The DataFrame from which to retrieve the value.
    column (str): The column name from which to retrieve the value.
    row (int): The row index from which to retrieve the value.
    datatype (type): The type to which the retrieved value should be converted.

    Returns:
    Datatype (Any|None): The value from the specified cell converted to the specified datatype.
         Returns None if the value is not found.
    """
    return (
        datatype(value=value)
        if (value := _get_value(data_frame, column, row))
        else None
    )


def _get_value_not_none(dataframe: pd.DataFrame, column: str, row: int) -> Any:
    """
    Retrieves the value from a specified column and row in a DataFrame, ensuring the value is not None.

    Parameters:
    dataframe (pd.DataFrame): The DataFrame from which to retrieve the value.
    column (str): The column name from which to retrieve the value.
    row (int): The row index from which to retrieve the value.

    Returns:
    Any: The value from the specified cell.

    Raises:
    AllotropeConversionError: If the value is None.
    """
    value = _get_value(dataframe, column, row)
    if value is None:
        msg = f"{constants.VALUE_ERROR} '{column}'."
        raise AllotropeConversionError(msg)
    return value


def _get_property_value_not_none(
    data_frame: pd.DataFrame, column: str, row: int, datatype: type
) -> DataType | None:
    """
    Retrieves the value from a specified column and row in a DataFrame and converts it
    to the specified datatype. If the value is not in the respective column it throws the Allotrope Exception.

    Parameters:
    data_frame (pd.DataFrame): The DataFrame from which to retrieve the value.
    column (str): The column name from which to retrieve the value.
    row (int): The row index from which to retrieve the value.
    datatype (type): The type to which the retrieved value should be converted.

    Returns:
    Datatype(Any): The value from the specified cell converted to the specified datatype.

    Raises:
    AllotropeConversionError: If the value is None.
    """
    return (
        datatype(value=value)
        if (value := _get_value_not_none(data_frame, column, row))
        else None
    )


class ThermoFisherQubitFlexParser(VendorParser):
    """
    A class provides the allotrope model of the Thermo Fisher Qubit Flex files
    """

    @property
    def display_name(self) -> str:
        """
        Provide the display name of the Adapter

        Returns: the Adapter display name as string
        """
        return constants.DISPLAY_NAME

    @property
    def release_state(self) -> ReleaseState:
        """
        Provide the Release state of the Adapter
        """
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        """
        Reads the content of the provided named file and returns it as Model

        Parameters:
        named_file_contents: contents of the provided named file

        Returns: the Model of the provided named file
        """
        return self._get_model(
            data=ThermoFisherQubitFlexReader.read(named_file_contents),
            filename=named_file_contents.original_file_name,
        )

    def _get_model(self, data: pd.DataFrame, filename: str) -> Model:
        """
        Reads the content of the provided named file and returns it as Model

        Parameters:
        data: dataframe of the provided named file
        filename: original filename of the provided named file

        Returns: the Model of the provided named file
        """
        software_version = _get_value(data, "Software Version", (len(data.index) - 1))
        return Model(
            field_asm_manifest="http://purl.allotrope.org/manifests/spectrophotometry/BENCHLING/2023/12/spectrophotometry.manifest",
            spectrophotometry_aggregate_document=SpectrophotometryAggregateDocument(
                spectrophotometry_document=self._get_spectrophotometry_document(data),
                data_system_document=DataSystemDocument(
                    file_name=filename,
                    software_name=constants.SOFTWARE_NAME,
                    software_version=software_version,
                    ASM_converter_name=self.get_asm_converter_name(),
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                device_system_document=DeviceSystemDocument(
                    model_number=constants.MODEL_NUMBER,
                    device_identifier=NOT_APPLICABLE,
                    brand_name=constants.BRAND_NAME,
                    product_manufacturer=constants.PRODUCT_MANUFACTURER,
                ),
            ),
        )

    def _get_spectrophotometry_document(
        self, data: pd.DataFrame
    ) -> list[SpectrophotometryDocumentItem]:
        """
        Reads the content of the provided named file and returns it as a list.

        Parameters:
        data: dataframe of the provided named file
        i: index of the dataframe

        Returns: the spectrophotometry document list
        """
        return [
            SpectrophotometryDocumentItem(
                measurement_aggregate_document=self._get_measurement_aggregate_document(
                    data, i
                )
            )
            for i in range(len(data.index))
        ]

    def _get_measurement_aggregate_document(
        self, data: pd.DataFrame, i: int
    ) -> MeasurementAggregateDocument:
        """
        Reads the content of the provided named file and returns it as a dictionary.

        Parameters:
        data: dataframe of the provided named file
        i: index of the dataframe

        Returns: the measurement aggregate document dictionary
        """
        return MeasurementAggregateDocument(
            measurement_time=self._get_date_time(
                str(_get_value_not_none(data, "Test Date", i))
            ),
            experiment_type=_get_value(data, "Assay Name", i),
            container_type=ContainerType.tube,
            measurement_document=self._get_measurement_document(data, i),
        )

    def _get_measurement_document(
        self, data: pd.DataFrame, i: int
    ) -> list[
        FluorescenceEmissionSpectrumDetectionMeasurementDocumentItems
        | FluorescencePointDetectionMeasurementDocumentItems
        | LuminescencePointDetectionMeasurementDocumentItems
        | UltravioletAbsorbancePointDetectionMeasurementDocumentItems
        | UltravioletAbsorbanceSpectrumDetectionMeasurementDocumentItems
    ]:
        """
        Reads the content of the provided named file and returns it as a dictionary.

        Parameters:
        data: dataframe of the provided named file
        i: index of the dataframe

        Returns: the measurement document dictionary
        """
        measurement_custom_document = {
            "reagent lot number": _get_value(data, "Reagent Lot#", i),
            "calibrated tubes": _get_value(data, "Calibrated Tubes", i),
        }
        # Check if all values in sample_custom_document are None
        if all(value is None for value in measurement_custom_document.values()):
            return [
                FluorescencePointDetectionMeasurementDocumentItems(
                    fluorescence=self._get_fluorescence_value(data, i),
                    measurement_identifier=random_uuid_str(),
                    sample_document=self._get_sample_document(data, i),
                    device_control_aggregate_document=self._get_device_control_document(
                        data, i
                    ),
                ),
            ]

        else:
            return [
                add_custom_information_document(
                    FluorescencePointDetectionMeasurementDocumentItems(
                        fluorescence=self._get_fluorescence_value(data, i),
                        measurement_identifier=random_uuid_str(),
                        sample_document=self._get_sample_document(data, i),
                        device_control_aggregate_document=self._get_device_control_document(
                            data, i
                        ),
                    ),
                    measurement_custom_document,
                )
            ]

    def _get_fluorescence_value(self, data: pd.DataFrame, i: int) -> Any:
        """
        Reads the content of the provided named file and returns the fluorescence value as string.

        Parameters:
        data: dataframe of the provided named file
        i: index of the dataframe

        Returns: the fluorescence value as string
        """
        value = _get_property_value_not_none(
            data, "Sample RFU", i, TQuantityValueRelativeFluorescenceUnit
        )
        return value

    def _get_sample_document(self, data: pd.DataFrame, i: int) -> SampleDocument:
        """
        Reads the content of the provided named file and returns it as a dictionary.

        Parameters:
        data: dataframe of the provided named file
        i: index of the dataframe

        Returns: the sample document dictionary
        """
        sample_id = _get_value(data, "Sample ID", i)
        location_id = _get_value(data, "Well", i)
        well_plate_id = _get_value(data, "Plate Barcode", i)
        if sample_id is None:
            sample_id = _get_value_not_none(data, "Sample Name", i)
        sample_custom_document = {
            "original sample concentration": _get_concentration_value(
                data, "Original Sample Conc.", "Original sample conc. units", i
            ),
            "qubit tube concentration": _get_concentration_value(
                data, "Qubit Tube Conc.", "Qubit tube conc. units", i
            ),
            "standard 1 concentration": _get_property_value(
                data, "Std 1 RFU", i, TQuantityValueRelativeFluorescenceUnit
            ),
            "standard 2 concentration": _get_property_value(
                data, "Std 2 RFU", i, TQuantityValueRelativeFluorescenceUnit
            ),
            "standard 3 concentration": _get_property_value(
                data, "Std 3 RFU", i, TQuantityValueRelativeFluorescenceUnit
            ),
            "last read standards": self._get_date_time(
                str(_get_value(data, "Test Date", i))
            ),
            "selected samples": _get_value(data, "Selected Samples", i),
        }
        if all(value is None for value in sample_custom_document.values()):
            return SampleDocument(
                sample_identifier=sample_id,
                batch_identifier=str(_get_value(data, "Run ID", i)),
                location_identifier=location_id,
                well_plate_identifier=well_plate_id,
            )
        else:
            return add_custom_information_document(
                SampleDocument(
                    sample_identifier=sample_id,
                    batch_identifier=str(_get_value(data, "Run ID", i)),
                    location_identifier=location_id,
                    well_plate_identifier=well_plate_id,
                ),
                sample_custom_document,
            )

    def _get_device_control_document(
        self, data: pd.DataFrame, i: int
    ) -> FluorescencePointDetectionDeviceControlAggregateDocument:
        """
        Reads the content of the provided named file and returns it as a dictionary.

        Parameters:
        data: dataframe of the provided named file
        i: index of the dataframe

        Returns: the device control document dictionary
        """
        custom_device_document = {
            "sample volume setting": _get_property_value(
                data, "Sample Volume (uL)", i, TQuantityValueMicroliter
            ),
            "operating minimum": _get_property_value(
                data, "Extended Low Range", i, TQuantityValueNanogramPerMicroliter
            ),
            "operating range": _get_property_value(
                data, "Core Range", i, TQuantityValueNanogramPerMicroliter
            ),
            "operating maximum": _get_property_value(
                data, "Extended High Range", i, TQuantityValueNanogramPerMicroliter
            ),
            "excitation setting": _get_value(data, "Excitation", i),
            "dilution factor": _get_property_value(
                data, "Dilution Factor", i, TQuantityValueUnitless
            ),
        }
        if all(value is None for value in custom_device_document.values()):
            return FluorescencePointDetectionDeviceControlAggregateDocument(
                device_control_document=[
                    FluorescencePointDetectionDeviceControlDocumentItem(
                        device_type=constants.DEVICE_TYPE
                    ),
                ]
            )
        else:
            return FluorescencePointDetectionDeviceControlAggregateDocument(
                device_control_document=[
                    add_custom_information_document(
                        FluorescencePointDetectionDeviceControlDocumentItem(
                            device_type=constants.DEVICE_TYPE
                        ),
                        custom_device_document,
                    )
                ]
            )



from __future__ import annotations

from collections.abc import Iterator
from io import StringIO
from re import search
from typing import Any, Literal

import chardet
import pandas as pd

from allotropy.allotrope.pandas_util import read_csv
from allotropy.constants import CHARDET_ENCODING, DEFAULT_ENCODING
from allotropy.exceptions import AllotropeConversionError, AllotropeParsingError
from allotropy.named_file_contents import NamedFileContents

EMPTY_STR_PATTERN = r"^\s*$"


def read_to_lines(named_file_contents: NamedFileContents) -> list[str]:
    stream_contents = named_file_contents.contents.read()
    encoding = named_file_contents.encoding
    raw_contents = (
        _decode(stream_contents, encoding)
        if isinstance(stream_contents, bytes)
        else stream_contents
    )
    contents = raw_contents.replace("\r\n", "\n")
    return contents.split("\n")


def determine_encoding(bytes_content: bytes, encoding: str | None) -> str:
    if not encoding:
        return DEFAULT_ENCODING
    if encoding != CHARDET_ENCODING:
        return encoding

    if not bytes_content:
        msg = "Unable to detect encoding for empty bytes string, file may be empty."
        raise AllotropeConversionError(msg)

    detected = chardet.detect(bytes_content)["encoding"]
    if not detected:
        msg = f"Unable to detect text encoding for file with content: {bytes_content!r}"
        raise AllotropeParsingError(msg)
    # Windows-1252 is a subset of UTF-8, and may lead to missing some data.
    if detected == "Windows-1252":
        detected = "utf-8"
    return detected


def _decode(bytes_content: bytes, encoding: str | None) -> str:
    encoding_to_use = determine_encoding(bytes_content, encoding)
    try:
        return bytes_content.decode(encoding_to_use)
    except UnicodeDecodeError as e:
        msg = f"Unable to decode bytes with encoding '{encoding_to_use}' with error: {e}, bytes: {bytes_content!r}"
        raise AllotropeParsingError(msg) from e
    except LookupError as e:
        msg = f"Invalid encoding: '{encoding}'."
        raise AllotropeConversionError(msg) from e


class LinesReader:
    lines: list[str]
    current_line: int

    def __init__(self, lines: list[str]) -> None:
        self.lines = lines
        self.current_line = 0

    def line_exists(self, line: int) -> bool:
        return 0 <= line < len(self.lines)

    def current_line_exists(self) -> bool:
        return self.line_exists(self.current_line)

    def get_line(self, line: int) -> str | None:
        return self.lines[line] if self.line_exists(line) else None

    def get(self) -> str | None:
        return self.lines[self.current_line] if self.current_line_exists() else None

    def match(self, match_pat: str) -> bool:
        line = self.get()
        return False if line is None else bool(search(match_pat, line))

    def is_empty(self, empty_pat: str = EMPTY_STR_PATTERN) -> bool:
        return self.match(empty_pat)

    def pop(self) -> str | None:
        line = self.get()
        if line is not None:
            self.current_line += 1
        return line

    def pop_if_match(self, match_pat: str) -> str | None:
        return self.pop() if self.match(match_pat) else None

    def pop_data(self) -> str | None:
        self.drop_empty()
        return self.pop()

    def drop_until(self, match_pat: str) -> str | None:
        while self.current_line_exists() and not self.match(match_pat):
            self.pop()
        return self.get()

    def drop_until_inclusive(self, match_pat: str) -> str | None:
        self.drop_until(match_pat)
        return self.pop()

    def drop_empty(self, empty_pat: str = EMPTY_STR_PATTERN) -> str | None:
        while self.current_line_exists() and self.is_empty(empty_pat):
            self.pop()
        return self.get()

    def drop_until_empty(self, empty_pat: str = EMPTY_STR_PATTERN) -> str | None:
        while self.current_line_exists() and not self.is_empty(empty_pat):
            self.pop()
        return self.get()

    def drop_until_empty_inclusive(
        self, empty_pat: str = EMPTY_STR_PATTERN
    ) -> str | None:
        self.drop_until_empty(empty_pat)
        return self.pop()

    def pop_until(self, match_pat: str) -> Iterator[str]:
        while self.current_line_exists() and not self.match(match_pat):
            line = self.pop()
            if line is not None:
                yield line

    def pop_until_empty(self, empty_pat: str = EMPTY_STR_PATTERN) -> Iterator[str]:
        while self.current_line_exists() and not self.is_empty(empty_pat):
            line = self.pop()
            if line is not None:
                yield line


class InvertedLinesReader(LinesReader):
    def __init__(self, lines: list[str]) -> None:
        self.lines = lines
        self.current_line = len(lines) - 1

    def pop(self) -> str | None:
        line = self.get()
        if line is not None:
            self.current_line -= 1
        return line


class CsvReader(LinesReader):
    def pop_csv_block_as_lines(self, empty_pat: str = EMPTY_STR_PATTERN) -> list[str]:
        self.drop_empty(empty_pat)
        lines = list(self.pop_until_empty(empty_pat))
        self.drop_empty(empty_pat)
        return lines

    def pop_csv_block_as_df(
        self,
        empty_pat: str = EMPTY_STR_PATTERN,
        header: int | Literal["infer"] | None = None,
        **kwargs: Any,
    ) -> pd.DataFrame | None:
        if lines := self.pop_csv_block_as_lines(empty_pat):
            return read_csv(
                StringIO("\n".join(lines)),
                dtype=None,
                # Prevent pandas from rounding decimal values, at the cost of some speed.
                float_precision="round_trip",
                header=header,
                **kwargs,
            )
        return None

    def drop_sections(self, match_pat: str) -> None:
        self.drop_empty()
        while self.match(match_pat):
            self.drop_until_empty()
            self.drop_empty()

    def pop_as_series(self, sep: str = " ") -> pd.Series[str] | None:
        line = self.pop()
        return None if line is None else pd.Series(line.split(sep))

    def lines_as_df(
        self,
        lines: list[str],
        header: int | Literal["infer"] | None = None,
        **kwargs: Any,
    ) -> pd.DataFrame | None:
        if lines:
            return read_csv(
                StringIO("\n".join(lines)),
                dtype=None,
                # Prevent pandas from rounding decimal values, at the cost of some speed.
                float_precision="round_trip",
                header=header,
                **kwargs,
            )
        return None



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path, PureWindowsPath
import re

from allotropy.allotrope.models.shared.definitions.definitions import JsonFloat
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    Data,
    ImageFeature,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ProcessedData,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.lines_reader import LinesReader
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    try_float_or_nan,
)


@dataclass(frozen=True)
class Plate:
    name: str
    wells: dict[str, JsonFloat]

    @staticmethod
    def create(reader: LinesReader) -> Plate:
        raw_name = assert_not_none(
            reader.pop(),
            msg="Unable to read name of assay data.",
        )

        name_match = assert_not_none(
            re.search("\t\t\t(.+)", raw_name),
            msg="Unable to parse name of assay data.",
        )
        name = re.sub(r"\s+", " ", name_match.group(1))

        if "ImmunoSpot Plate Code" in name:
            name = "Spot Count"

        raw_columns = assert_not_none(
            reader.pop(),
            msg=f"Unable to read data from {name}",
        )
        columns = re.sub(r"\s+", " ", raw_columns.strip())

        reader.drop_empty()

        wells = {}
        for raw_line in reader.pop_until_empty():
            line = re.sub(r"\s+", " ", raw_line)
            if match := re.match(r"^([A-Z])\s+(.+)", line):
                raw_values = match.group(2).strip()
                for column, value in zip(
                    columns.split(), raw_values.split(), strict=True
                ):
                    wells[match.group(1) + column] = try_float_or_nan(value)

        return Plate(name, wells)


@dataclass(frozen=True)
class AssayData:
    plates: list[Plate]
    identifier: str | None

    @staticmethod
    def create(reader: LinesReader) -> AssayData:
        reader.drop_until_inclusive("Unprocessed Data$")

        reader.drop_empty()
        plate_code_line = reader.get() or ""
        assert_not_none(
            re.search("Plate Code =", plate_code_line),
            msg="Unable to find ImmunoSpot Plate Code line",
        )

        identifier = (
            match.group(1)
            if (match := re.search(r"Plate Code = ([\w ]+)", plate_code_line))
            else None
        )

        plates = []
        while reader.current_line_exists():
            reader.drop_empty()
            plates.append(Plate.create(reader))
            reader.drop_empty()

        if not plates:
            msg = "Unable to find plate information."
            raise AllotropeConversionError(msg)

        return AssayData(plates, identifier)


def _create_measurement_groups(
    assay_data: AssayData, metadata: Metadata
) -> list[MeasurementGroup]:
    well_plate_identifier = (
        assay_data.identifier or Path(assert_not_none(metadata.file_name)).stem
    )
    plate_well_count = len(assay_data.plates[0].wells)
    return [
        MeasurementGroup(
            plate_well_count=plate_well_count,
            measurements=[
                Measurement(
                    type_=MeasurementType.OPTICAL_IMAGING,
                    identifier=random_uuid_str(),
                    well_plate_identifier=well_plate_identifier,
                    location_identifier=well_position,
                    sample_identifier=f"{well_plate_identifier}_{well_position}",
                    processed_data=ProcessedData(
                        random_uuid_str(),
                        features=[
                            ImageFeature(
                                identifier=random_uuid_str(),
                                feature=plate.name,
                                result=plate.wells[well_position],
                            )
                            for plate in assay_data.plates
                        ],
                    ),
                )
            ],
        )
        for well_position in assay_data.plates[0].wells
    ]


def _create_metadata(reader: LinesReader) -> Metadata:
    data = {}
    for i in range(3):  # for first three lines
        line = assert_not_none(
            reader.pop(),
            msg=f"Unable to read line number {i+1}",
        )
        for element in line.split(";"):
            parts = element.split(":", maxsplit=1)
            if len(parts) > 1:
                data[parts[0].strip()] = parts[1].strip()

    raw_path = assert_not_none(
        reader.pop(),
        msg="Unable to read file path",
    )

    path = PureWindowsPath(raw_path)

    reader.drop_until_empty()
    reader.drop_empty()

    analyzer_serial_number = assert_not_none(
        data.get("Analyzer Serial number"),
        msg="Unable to find analyzer serial number.",
    )
    software_info = assert_not_none(
        data.get("Software version"),
        msg="Unable to find software version",
    )
    return Metadata(
        device_identifier=NOT_APPLICABLE,
        device_type="imager",
        detection_type="optical-imaging",
        model_number=assert_not_none(
            re.match(r"^(\w+)-(\w+)", analyzer_serial_number),
            msg="Unable to parse analyzer serial number.",
        ).group(1),
        equipment_serial_number=analyzer_serial_number,
        data_system_instance_id=assert_not_none(
            data.get("Computer name"),
            msg="Unable to find computer name",
        ),
        file_name=path.name,
        unc_path=raw_path,
        product_manufacturer="CTL",
        software_name="ImmunoSpot",
        software_version=assert_not_none(
            re.match(r"^ImmunoSpot ([\d\.]+)$", software_info),
            msg="Unable to parse software version",
        ).group(1),
        measurement_time=assert_not_none(
            data.get("Counted"),
            msg="Unable to find counted timestamp.",
        ),
        analyst=assert_not_none(
            data.get("Authenticated user"),
            msg="Unable to find authenticated user.",
        ),
    )


def create_data(reader: LinesReader) -> Data:
    metadata = _create_metadata(reader)

    reader.drop_empty()
    reader.drop_until_empty()  # ignore assay info
    reader.drop_empty()

    assay_data = AssayData.create(reader)
    return Data(
        metadata=metadata,
        measurement_groups=_create_measurement_groups(assay_data, metadata),
    )



from allotropy.allotrope.models.adm.cell_counting.benchling._2023._11.cell_counting import (
    CellCountingAggregateDocument,
    CellCountingDetectorDeviceControlAggregateDocument,
    CellCountingDetectorMeasurementDocumentItem,
    CellCountingDocumentItem,
    DataProcessingDocument,
    DataSystemDocument,
    DeviceControlDocumentItemModel,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    Model,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueMicrometer,
    TQuantityValueMillionCellsPerMilliliter,
    TQuantityValuePercent,
    TQuantityValueUnitless,
)
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.chemometec_nucleoview.constants import (
    NUCLEOCOUNTER_DETECTION_TYPE,
    NUCLEOCOUNTER_DEVICE_TYPE,
    NUCLEOCOUNTER_SOFTWARE_NAME,
)
from allotropy.parsers.chemometec_nucleoview.nucleoview_reader import NucleoviewReader
from allotropy.parsers.chemometec_nucleoview.nucleoview_structure import Row
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import quantity_or_none
from allotropy.parsers.vendor_parser import VendorParser


class ChemometecNucleoviewParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "ChemoMetec Nucleoview"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        contents = named_file_contents.contents
        filename = named_file_contents.original_file_name
        rows = Row.create_rows(NucleoviewReader.read(contents))
        return self._get_model(rows, filename)

    def _get_model(self, rows: list[Row], filename: str) -> Model:
        return Model(
            field_asm_manifest="http://purl.allotrope.org/manifests/cell-counting/BENCHLING/2023/11/cell-counting.manifest",
            cell_counting_aggregate_document=CellCountingAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    model_number=rows[0].model_number,
                    equipment_serial_number=rows[0].equipment_serial_number,
                ),
                data_system_document=DataSystemDocument(
                    file_name=filename,
                    software_name=NUCLEOCOUNTER_SOFTWARE_NAME,
                    ASM_converter_name=self.get_asm_converter_name(),
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                    software_version=rows[0].software_version,
                ),
                cell_counting_document=[
                    self._get_cell_counting_document_item(row) for row in rows
                ],
            ),
        )

    def _get_cell_counting_document_item(self, row: Row) -> CellCountingDocumentItem:
        return CellCountingDocumentItem(
            analyst=row.analyst,
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_document=[
                    CellCountingDetectorMeasurementDocumentItem(
                        measurement_identifier=random_uuid_str(),
                        measurement_time=self._get_date_time(row.timestamp),
                        sample_document=SampleDocument(
                            sample_identifier=row.sample_identifier
                        ),
                        device_control_aggregate_document=CellCountingDetectorDeviceControlAggregateDocument(
                            device_control_document=[
                                DeviceControlDocumentItemModel(
                                    device_type=NUCLEOCOUNTER_DEVICE_TYPE,
                                    detection_type=NUCLEOCOUNTER_DETECTION_TYPE,
                                )
                            ]
                        ),
                        processed_data_aggregate_document=ProcessedDataAggregateDocument(
                            processed_data_document=[
                                ProcessedDataDocumentItem(
                                    data_processing_document=DataProcessingDocument(
                                        cell_density_dilution_factor=quantity_or_none(
                                            TQuantityValueUnitless,
                                            row.multiplication_factor,
                                        )
                                    ),
                                    viability__cell_counter_=TQuantityValuePercent(
                                        value=row.viability_percent
                                    ),
                                    viable_cell_density__cell_counter_=TQuantityValueMillionCellsPerMilliliter(
                                        value=row.live_cell_count
                                    ),
                                    dead_cell_density__cell_counter_=quantity_or_none(
                                        TQuantityValueMillionCellsPerMilliliter,
                                        row.dead_cell_count,
                                    ),
                                    total_cell_density__cell_counter_=quantity_or_none(
                                        TQuantityValueMillionCellsPerMilliliter,
                                        row.total_cell_count,
                                    ),
                                    average_total_cell_diameter=quantity_or_none(
                                        TQuantityValueMicrometer,
                                        row.estimated_cell_diameter,
                                    ),
                                )
                            ]
                        ),
                    )
                ],
            ),
        )



BRAND_NAME = "TapeStation"
DETECTION_TYPE = "fluorescence"
DEVICE_TYPE = "electrophoresis device"
PRODUCT_MANUFACTURER = "Agilent"
SOFTWARE_NAME = "TapeStation Analysis Software"

UNIT_CLASS_LOOKUP: dict[str, str] = {
    "nt": "#",
    "bp": "#",
    "kD": "kDa",
}
NON_CALCULATED_DATA_TAGS_SAMPLE = [
    "Comment",
    "Observations",
    "Peaks",
    "Regions",
    "ScreenTapeID",
    "WellNumber",
]
NON_CALCULATED_DATA_TAGS_PEAK = [
    "Area",
    "Comment",
    "FromMW",
    "Height",
    "Number",
    "Observations",
    "PercentIntegratedArea",
    "PercentOfTotal",
    "Size",
    "ToMW",
]
NON_CALCULATED_DATA_TAGS_REGION = [
    "From",
    "To",
    "Area",
    "PercentOfTotal",
    "Comment",
]



from __future__ import annotations

from pathlib import PureWindowsPath

import numpy as np
import pandas as pd

from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.values import (
    assert_not_empty_df,
    assert_not_none,
)


class DesignQuantstudioContents:
    @staticmethod
    def create(named_file_contents: NamedFileContents) -> DesignQuantstudioContents:
        if not named_file_contents.original_file_name.endswith("xlsx"):
            extension = PureWindowsPath(named_file_contents.original_file_name).suffix
            msg = f"Invalid file extension for AppBio QuantStudio Design & Analysis: '{extension}', must be 'xlsx'"
            raise AllotropeConversionError(msg)
        raw_contents = pd.read_excel(  # type: ignore[call-overload]
            named_file_contents.contents,
            header=None,
            sheet_name=None,
            engine="calamine",
        )
        return DesignQuantstudioContents(raw_contents)

    def __init__(self, raw_contents: dict[str, pd.DataFrame]) -> None:
        contents = {
            str(name): df.replace(np.nan, None) for name, df in raw_contents.items()
        }

        self.header = self._get_header(contents)
        self.data = self._get_data(
            contents,
        )

    def _get_header_size(self, sheet: pd.DataFrame) -> int:
        # Find the first blank line
        for idx, * (title, *_) in sheet.itertuples():
            if title is None:
                return int(idx)
        msg = "Invalid file format, expected a blank line indicating the end of the header section."
        raise AllotropeConversionError(msg)

    def _get_header(self, contents: dict[str, pd.DataFrame]) -> SeriesData:
        sheet = assert_not_none(
            contents.get("Results"),
            msg="Unable to find 'Results' sheet.",
        )

        header_size = self._get_header_size(sheet)

        data = {}
        for _, * (title, value, *_) in sheet.iloc[:header_size].itertuples():
            data[str(title)] = None if value is None else str(value)

        header = pd.Series(data)
        header.index = header.index.str.strip()
        return SeriesData(header)

    def _get_data(self, contents: dict[str, pd.DataFrame]) -> dict[str, pd.DataFrame]:
        data_structure = {}
        for name, sheet in contents.items():
            # header_size + 1 for empty line between header and data
            header_size = self._get_header_size(sheet) + 1
            data = sheet.iloc[header_size:].reset_index(drop=True)
            data.columns = pd.Index(data.iloc[0])
            data_structure[name] = data.drop(0)
        return data_structure

    def has_sheet(self, sheet_name: str) -> bool:
        return sheet_name in self.data

    def get_sheet_or_none(self, sheet_name: str) -> pd.DataFrame | None:
        return self.data.get(sheet_name)

    def get_non_empty_sheet_or_none(self, sheet_name: str) -> pd.DataFrame | None:
        sheet = self.get_sheet_or_none(sheet_name)
        return None if sheet is None or sheet.empty else sheet

    def get_sheet(self, sheet_name: str) -> pd.DataFrame:
        return assert_not_none(
            self.get_sheet_or_none(sheet_name),
            msg=f"Unable to find '{sheet_name}' sheet in file.",
        )

    def get_non_empty_sheet(self, sheet_name: str) -> pd.DataFrame:
        return assert_not_empty_df(
            self.get_sheet(sheet_name),
            msg=f"sheet '{sheet_name}' is empty.",
        )



# mypy: disallow_any_generics = False

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass

from dateutil import parser
import pandas as pd

from allotropy.allotrope.models.shared.definitions.definitions import JsonFloat, NaN
from allotropy.exceptions import AllotropyParserError
from allotropy.parsers.roche_cedex_bioht.constants import (
    BELOW_TEST_RANGE,
    MAX_MEASUREMENT_TIME_GROUP_DIFFERENCE,
)
from allotropy.parsers.roche_cedex_bioht.roche_cedex_bioht_reader import (
    RocheCedexBiohtReader,
)
from allotropy.parsers.utils.pandas import map_rows, SeriesData


@dataclass(frozen=True)
class Title:
    data_processing_time: str
    analyst: str
    device_serial_number: str
    model_number: str | None = None
    software_version: str | None = None

    @staticmethod
    def create(title_data: SeriesData) -> Title:
        return Title(
            title_data[str, "data processing time"],
            title_data[str, "analyst"],
            title_data[str, "device serial number"],
            title_data.get(str, "model number"),
            title_data.get(str, "software version"),
        )


@dataclass(frozen=True)
class Measurement:
    name: str
    measurement_time: str
    concentration_value: JsonFloat
    unit: str
    error: str | None = None

    @staticmethod
    def create(data: SeriesData) -> Measurement:
        error = data.get(str, "flag", "").strip()
        # TODO: handle other errors
        if BELOW_TEST_RANGE not in error:
            error = ""
        concentration_value = (
            NaN
            if BELOW_TEST_RANGE in error
            else data.get(float, "concentration value", NaN)
        )
        return Measurement(
            data[str, "analyte name"],
            data[str, "measurement time"],
            concentration_value,
            data[str, "concentration unit"],
            error or None,
        )


def create_measurements(data: pd.DataFrame) -> dict[str, dict[str, Measurement]]:
    measurements = sorted(
        map_rows(data, Measurement.create), key=lambda a: a.measurement_time
    )

    # Dict from measurement time to data
    groups: defaultdict[str, dict[str, Measurement]] = defaultdict(dict)

    current_measurement_time = measurements[0].measurement_time
    previous_measurement_time = current_measurement_time
    for analyte in measurements:
        time_diff = parser.parse(analyte.measurement_time) - parser.parse(
            previous_measurement_time
        )
        if time_diff > MAX_MEASUREMENT_TIME_GROUP_DIFFERENCE:
            current_measurement_time = analyte.measurement_time
        if analyte.name in groups[current_measurement_time]:
            if analyte.concentration_value is NaN:
                continue
            # NOTE: if this fails, it's probably because MAX_MEASUREMENT_TIME_GROUP_DIFFERENCE is too big
            # and we're erroneously grouping two groups of measurements into one.
            # We could potentially make this more robust by just splitting into a new group if a duplicate
            # measurement is found, but cross that bridge when we come to it.
            if (
                groups[current_measurement_time][analyte.name].concentration_value
                is not NaN
            ):
                msg = f"Duplicate measurement for {analyte.name} in the same measurement group: {analyte.concentration_value} vs {groups[current_measurement_time][analyte.name].concentration_value}"
                raise AllotropyParserError(msg)
        groups[current_measurement_time][analyte.name] = analyte
        previous_measurement_time = analyte.measurement_time

    return dict(groups)


@dataclass(frozen=True)
class Sample:
    name: str
    measurements: dict[str, dict[str, Measurement]]
    batch: str | None = None

    @staticmethod
    def create(name: str, batch: str | None, sample_data: pd.DataFrame) -> Sample:
        return Sample(
            name,
            create_measurements(sample_data),
            batch=batch or None,
        )


@dataclass(frozen=True)
class Data:
    title: Title
    samples: list[Sample]

    @staticmethod
    def create(reader: RocheCedexBiohtReader) -> Data:
        return Data(
            title=Title.create(reader.title_data),
            samples=[
                Sample.create(name, batch, samples_data)
                for (name, batch), samples_data in reader.samples_data.groupby(
                    # A sample group is defined by both the sample and the batch identifier
                    ["sample identifier", "batch identifier"]
                )
            ],
        )



from collections.abc import Collection
from enum import Enum
from typing import Any, TypeVar


# Base exception for errors raised by allotropy.
class AllotropyError(Exception):
    ...


# Unexpected error when reading input data.
class AllotropeParsingError(AllotropyError):
    ...


# Unexpected error when validating output against schema.
class AllotropeValidationError(AllotropyError):
    ...


# Unexpected error when converting allotropy dataclass model to json.
class AllotropeSerializationError(AllotropyError):
    ...


# Expected error caused by a programming error, indicating a bug.
class AllotropyParserError(AllotropyError):
    ...


# Expected error caused by bad input data, with a message telling user what the problem is.
class AllotropeConversionError(AllotropyError):
    ...


def list_values(values: Collection[Any] | type[Enum]) -> list[str]:
    return sorted([str(v.value if isinstance(v, Enum) else v) for v in values])


T = TypeVar("T")


def valid_value_or_raise(
    name: str, values: set[T], valid_values: Collection[T] | type[Enum]
) -> T:
    if len(values) == 1:
        return values.pop()
    msg = f"Could not infer {name}, expecting exactly one of {list_values(valid_values)}, found {list_values(values)}"
    raise AllotropeConversionError(msg)


def get_key_or_error(name: str, key: str, mapping: dict[str, T]) -> T:
    try:
        return mapping[key]
    except KeyError as e:
        msg = msg_for_error_on_unrecognized_value(name, key, mapping.keys())
        raise AllotropeConversionError(msg) from e


def msg_for_error_on_unrecognized_value(
    name: str, value: T, valid_values: Collection[T] | type[Enum] | None = None
) -> str:
    msg = f"Unrecognized {name}: '{value}'."
    if valid_values:
        msg += f" Expecting one of {list_values(valid_values)}."
    return msg



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterator
from dataclasses import dataclass
from enum import Enum
import re

import pandas as pd

from allotropy.allotrope.models.shared.definitions.definitions import JsonFloat
from allotropy.exceptions import (
    AllotropeConversionError,
    get_key_or_error,
)
from allotropy.parsers.lines_reader import CsvReader
from allotropy.parsers.utils.pandas import rm_df_columns, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    num_to_chars,
    try_float,
    try_float_or_nan,
    try_float_or_none,
    try_int,
    try_int_or_none,
    try_non_nan_float_or_none,
)

BLOCKS_LINE_REGEX = r"^##BLOCKS=\s*(\d+)$"
END_LINE_REGEX = "~End"
EXPORT_VERSION = "1.3"


NUM_WELLS_TO_PLATE_DIMENSIONS: dict[int, tuple[int, int]] = {
    6: (3, 2),
    12: (4, 3),
    24: (6, 4),
    48: (8, 6),
    96: (12, 8),
    384: (24, 16),
    1536: (48, 32),
}


def num_wells_to_n_columns(well_count: int) -> int:
    if dimensions := NUM_WELLS_TO_PLATE_DIMENSIONS.get(well_count):
        return dimensions[0]

    num_wells = ",".join(
        [str(num_wells) for num_wells in NUM_WELLS_TO_PLATE_DIMENSIONS]
    )
    msg = (
        f"Unknown number of wells '{well_count}'. Only accepted values are {num_wells}"
    )
    raise AllotropeConversionError(msg)


class ReadType(Enum):
    ENDPOINT = "Endpoint"
    KINETIC = "Kinetic"
    SPECTRUM = "Spectrum"
    WELL_SCAN = "Well Scan"


class ExportFormat(Enum):
    TIME_FORMAT = "TimeFormat"
    PLATE_FORMAT = "PlateFormat"


class DataType(Enum):
    RAW = "Raw"
    REDUCED = "Reduced"
    BOTH = "Both"


class ScanPosition(Enum):
    BOTTOM = "Bottom"
    TOP = "Top"
    NONE = None


@dataclass(frozen=True)
class Block:
    block_type: str


@dataclass(frozen=True)
class GroupDataElementEntry:
    name: str
    value: float


@dataclass(frozen=True)
class GroupDataElement:
    sample: str
    position: str
    plate: str
    entries: list[GroupDataElementEntry]


@dataclass(frozen=True)
class GroupSampleData:
    identifier: str
    data_elements: list[GroupDataElement]
    aggregated_entries: list[GroupDataElementEntry]

    @staticmethod
    def create(data: pd.DataFrame) -> GroupSampleData:
        row_data = [SeriesData(row) for _, row in data.iterrows()]
        top_row = row_data[0]
        identifier = top_row[str, "Sample"]
        data = rm_df_columns(data, r"^Sample$|^Standard Value|^R$|^Unnamed: \d+$")
        numeric_columns = [
            column
            for column in data.columns
            if top_row.get(float, column, validate=SeriesData.NOT_NAN) is not None
        ]

        normal_columns = []
        aggregated_columns = []
        for column in numeric_columns:
            if data[column].iloc[1:].isnull().all():
                aggregated_columns.append(column)
            else:
                normal_columns.append(column)

        return GroupSampleData(
            identifier=identifier,
            data_elements=[
                GroupDataElement(
                    sample=identifier,
                    position=row[str, ["Well", "Wells"]],
                    plate=row[str, "WellPlateName"],
                    entries=[
                        GroupDataElementEntry(
                            name=column_name, value=row[float, column_name]
                        )
                        for column_name in normal_columns
                    ],
                )
                for row in row_data
            ],
            aggregated_entries=[
                GroupDataElementEntry(
                    name=column_name,
                    value=top_row[float, column_name],
                )
                for column_name in aggregated_columns
            ],
        )


@dataclass(frozen=True)
class GroupData:
    name: str
    sample_data: list[GroupSampleData]

    @staticmethod
    def create(reader: CsvReader) -> GroupData:
        name = assert_not_none(
            reader.pop(),
            msg="Unable to find group block name.",
        ).removeprefix("Group: ")

        with pd.option_context("future.no_silent_downcasting", True):  # noqa: FBT003
            data = assert_not_none(
                reader.pop_csv_block_as_df(sep="\t", header=0),
                msg="Unable to find group block data.",
            ).replace(r"^\s+$", None, regex=True)

        assert_not_none(
            data.get("Sample"),
            msg=f"Unable to find sample identifier column in group data {name}",
        )

        samples = data["Sample"].ffill()
        return GroupData(
            name=name,
            sample_data=[
                GroupSampleData.create(data.iloc[sample_entries.index])
                for _, sample_entries in samples.groupby(samples)
            ],
        )


@dataclass(frozen=True)
class GroupColumns:
    data: dict[str, str]

    @staticmethod
    def create(reader: CsvReader) -> GroupColumns:
        data = assert_not_none(
            reader.pop_csv_block_as_df(sep="\t", header=0),
            msg="Unable to find group block columns.",
        )

        if "Formula Name" not in data:
            msg = "Unable to find 'Formula Name' in group block columns."
            raise AllotropeConversionError(msg)

        if "Formula" not in data:
            msg = "Unable to find 'Formula' in group block columns."
            raise AllotropeConversionError(msg)

        return GroupColumns(
            data=dict(zip(data["Formula Name"], data["Formula"], strict=True)),
        )


@dataclass(frozen=True)
class GroupSummaries:
    data: list[str]

    @staticmethod
    def create(reader: CsvReader) -> GroupSummaries:
        data = list(reader.pop_until_empty())
        reader.drop_empty()
        return GroupSummaries(data)


@dataclass(frozen=True)
class GroupBlock(Block):
    group_data: GroupData
    group_columns: GroupColumns
    group_summaries: GroupSummaries

    @staticmethod
    def create(reader: CsvReader) -> GroupBlock:
        return GroupBlock(
            block_type="Group",
            group_data=GroupData.create(reader),
            group_columns=GroupColumns.create(reader),
            group_summaries=GroupSummaries.create(reader),
        )


# TODO do we need to do anything with these?
@dataclass(frozen=True)
class NoteBlock(Block):
    pass


@dataclass(frozen=True)
class PlateHeader:
    name: str
    export_version: str
    export_format: str
    read_type: str
    data_type: str
    kinetic_points: int
    num_wavelengths: int
    wavelengths: list[float]
    num_columns: int
    num_wells: int
    concept: str
    read_mode: str
    unit: str
    scan_position: ScanPosition
    reads_per_well: float
    pmt_gain: str | None
    num_rows: int
    excitation_wavelengths: list[int] | None
    cutoff_filters: list[int] | None


@dataclass
class DataElement:
    uuid: str
    plate: str
    temperature: float | None
    wavelength: float
    position: str
    value: JsonFloat
    sample_id: str | None = None

    @property
    def sample_identifier(self) -> str:
        return self.sample_id if self.sample_id else f"{self.plate} {self.position}"


@dataclass(frozen=True)
class ReducedDataElement:
    position: str
    value: float


@dataclass(frozen=True)
class PlateWavelengthData:
    wavelength: float
    data_elements: dict[str, DataElement]

    @staticmethod
    def create(
        plate_name: str,
        temperature: float | None,
        wavelength: float,
        df_data: pd.DataFrame,
    ) -> PlateWavelengthData:
        data = {
            f"{num_to_chars(row_idx)}{col}": value
            for row_idx, *row_data in df_data.itertuples()
            for col, value in zip(df_data.columns, row_data, strict=True)
        }
        return PlateWavelengthData(
            wavelength,
            data_elements={
                str(position): DataElement(
                    uuid=random_uuid_str(),
                    plate=plate_name,
                    temperature=temperature,
                    wavelength=wavelength,
                    position=str(position),
                    value=try_float_or_nan(value),
                )
                for position, value in data.items()
            },
        )


@dataclass(frozen=True)
class PlateKineticData:
    temperature: float | None
    wavelength_data: list[PlateWavelengthData]

    @staticmethod
    def create(
        reader: CsvReader,
        header: PlateHeader,
        columns: pd.Series[str],
    ) -> PlateKineticData:

        # use plate dimensions to determine how many rows of plate block to read
        dimensions = assert_not_none(
            NUM_WELLS_TO_PLATE_DIMENSIONS.get(header.num_wells),
            msg="unable to determine plate dimensions",
        )
        rows = dimensions[1]
        lines = []
        # read number of rows in plate
        for _ in range(rows):
            lines.append(reader.pop() or "")
        reader.drop_empty()

        # convert rows to df
        data = assert_not_none(
            reader.lines_as_df(lines=lines, sep="\t"),
            msg="unable to find data from plate block.",
        )
        data.columns = pd.Index(columns)

        # get temperature from the first column of the first row with value
        temperature = try_float_or_none(
            str(data.iloc[int(pd.to_numeric(data.first_valid_index())), 1])
        )

        return PlateKineticData(
            temperature=temperature,
            wavelength_data=PlateKineticData._get_wavelength_data(
                header.name,
                temperature,
                header,
                data.iloc[:, 2:],
            ),
        )

    @staticmethod
    def _get_wavelength_data(
        plate_name: str,
        temperature: float | None,
        header: PlateHeader,
        w_data: pd.DataFrame,
    ) -> list[PlateWavelengthData]:
        wavelength_data = []
        for idx in range(header.num_wavelengths):
            wavelength = header.wavelengths[idx]
            start = idx * (header.num_columns + 1)
            end = start + header.num_columns
            wavelength_data.append(
                PlateWavelengthData.create(
                    plate_name,
                    temperature,
                    wavelength,
                    w_data.iloc[:, start:end],
                )
            )
        return wavelength_data


@dataclass(frozen=True)
class PlateRawData:
    kinetic_data: list[PlateKineticData]

    @staticmethod
    def create(reader: CsvReader, header: PlateHeader) -> PlateRawData:
        columns = assert_not_none(
            reader.pop_as_series(sep="\t"),
            msg="unable to find data columns for plate block raw data.",
        )

        return PlateRawData(
            kinetic_data=[
                PlateKineticData.create(reader, header, columns)
                for _ in range(header.kinetic_points)
            ]
        )


@dataclass(frozen=True)
class PlateReducedData:
    data: list[ReducedDataElement]

    @staticmethod
    def create(reader: CsvReader, header: PlateHeader) -> PlateReducedData:
        raw_data = assert_not_none(
            reader.pop_csv_block_as_df(sep="\t", header=0),
            msg="Unable to find reduced data for plate block.",
        )

        start = 2
        df_data = raw_data.iloc[:, start : (start + header.num_columns)]

        reduced_data_elements = []
        for row, *data in df_data.itertuples():
            for col, str_value in zip(df_data.columns, data, strict=True):
                value = try_non_nan_float_or_none(str_value)
                if value is not None:
                    reduced_data_elements.append(
                        ReducedDataElement(
                            position=f"{num_to_chars(row)}{col}",
                            value=value,
                        )
                    )
        return PlateReducedData(data=reduced_data_elements)


@dataclass(frozen=True)
class PlateData:
    raw_data: PlateRawData
    reduced_data: PlateReducedData | None

    @staticmethod
    def create(
        reader: CsvReader,
        header: PlateHeader,
    ) -> PlateData:
        return PlateData(
            raw_data=PlateRawData.create(reader, header),
            reduced_data=(
                PlateReducedData.create(reader, header)
                if reader.current_line_exists()
                else None
            ),
        )

    def iter_data_elements(self, position: str) -> Iterator[DataElement]:
        for kinetic_data in self.raw_data.kinetic_data:
            for wavelength_data in kinetic_data.wavelength_data:
                yield wavelength_data.data_elements[position]


@dataclass(frozen=True)
class TimeKineticData:
    temperature: float | None
    data_elements: dict[str, DataElement]

    @staticmethod
    def create(
        plate_name: str,
        wavelength: float,
        row: pd.Series[float],
    ) -> TimeKineticData:
        temperature = try_non_nan_float_or_none(str(row.iloc[1]))

        return TimeKineticData(
            temperature=temperature,
            data_elements={
                str(position): DataElement(
                    uuid=random_uuid_str(),
                    plate=plate_name,
                    temperature=temperature,
                    wavelength=wavelength,
                    position=str(position),
                    value=try_float_or_nan(str(value)),
                )
                for position, value in row.iloc[2:].items()
            },
        )


@dataclass(frozen=True)
class TimeWavelengthData:
    wavelength: float
    kinetic_data: list[TimeKineticData]

    @staticmethod
    def create(
        reader: CsvReader,
        plate_name: str,
        wavelength: float,
        columns: pd.Series[str],
    ) -> TimeWavelengthData:
        data = assert_not_none(
            reader.pop_csv_block_as_df(sep="\t"),
            msg="unable to find raw data from time block.",
        )
        data.columns = pd.Index(columns)
        return TimeWavelengthData(
            wavelength=wavelength,
            kinetic_data=[
                TimeKineticData.create(plate_name, wavelength, row)
                for _, row in data.iterrows()
            ],
        )


@dataclass(frozen=True)
class TimeRawData:
    wavelength_data: list[TimeWavelengthData]

    @staticmethod
    def create(reader: CsvReader, header: PlateHeader) -> TimeRawData:
        columns = assert_not_none(
            reader.pop_as_series(sep="\t"),
            msg="unable to find data columns for time block raw data.",
        )

        return TimeRawData(
            wavelength_data=[
                TimeWavelengthData.create(
                    reader,
                    header.name,
                    wavelength,
                    columns,
                )
                for wavelength in header.wavelengths
            ]
        )


@dataclass(frozen=True)
class TimeReducedData:
    data: list[ReducedDataElement]

    @staticmethod
    def create(reader: CsvReader, header: PlateHeader) -> TimeReducedData:
        columns = assert_not_none(
            reader.pop_as_series(sep="\t"),
            msg="unable to find columns for time block reduced data.",
        )
        data = assert_not_none(
            reader.pop_as_series(sep="\t"),
            msg="unable to find reduced data from time block.",
        )
        data.index = pd.Index(columns)

        reduced_data_elements = []
        for pos, str_value in data[2 : header.num_wells + 2].items():
            value = try_non_nan_float_or_none(str_value)
            if value is not None:
                reduced_data_elements.append(
                    ReducedDataElement(
                        position=str(pos),
                        value=value,
                    )
                )
        return TimeReducedData(data=reduced_data_elements)


@dataclass(frozen=True)
class TimeData:
    raw_data: TimeRawData
    reduced_data: TimeReducedData | None

    @staticmethod
    def create(
        reader: CsvReader,
        header: PlateHeader,
    ) -> TimeData:
        return TimeData(
            raw_data=TimeRawData.create(reader, header),
            reduced_data=(
                TimeReducedData.create(reader, header)
                if reader.current_line_exists()
                else None
            ),
        )

    def iter_data_elements(self, position: str) -> Iterator[DataElement]:
        for wavelength_data in self.raw_data.wavelength_data:
            for kinetic_data in wavelength_data.kinetic_data:
                yield kinetic_data.data_elements[position]


@dataclass(frozen=True)
class PlateBlock(ABC, Block):
    header: PlateHeader
    block_data: PlateData | TimeData

    @staticmethod
    def read_header(reader: CsvReader) -> pd.Series[str]:
        raw_header_series = assert_not_none(
            reader.pop_as_series(sep="\t"),
            msg="Unable to find plate block header.",
        )
        return raw_header_series.replace("", None).str.strip()

    @staticmethod
    def get_plate_block_cls(header_series: pd.Series[str]) -> type[PlateBlock]:
        plate_block_cls: dict[str, type[PlateBlock]] = {
            "Absorbance": AbsorbancePlateBlock,
            "Fluorescence": FluorescencePlateBlock,
            "Luminescence": LuminescencePlateBlock,
        }
        read_mode = header_series[5]
        return get_key_or_error("read mode", read_mode, plate_block_cls)

    @staticmethod
    @abstractmethod
    def get_plate_block_type() -> str:
        ...

    @classmethod
    def parse_header(cls, header: pd.Series[str]) -> PlateHeader:
        raise NotImplementedError

    @classmethod
    def check_export_version(cls, export_version: str) -> None:
        if export_version != EXPORT_VERSION:
            msg = f"Unsupported export version {export_version}; only {EXPORT_VERSION} is supported."
            raise AllotropeConversionError(msg)

    @classmethod
    def check_read_type(cls, read_type: str) -> None:
        if read_type != ReadType.ENDPOINT.value:
            msg = f"Only Endpoint measurements can be processed at this time, got: {read_type}"
            raise AllotropeConversionError(msg)

    @classmethod
    def check_data_type(cls, data_type: str) -> None:
        if data_type not in (DataType.RAW.value, DataType.BOTH.value):
            msg = f"The SoftMax Pro file is required to include either 'Raw' or 'Both' (Raw and Reduced) data for all plates, got {data_type}."
            raise AllotropeConversionError(msg)

    @classmethod
    def check_num_wavelengths(
        cls, wavelengths: list[float], num_wavelengths: int
    ) -> None:
        if len(wavelengths) != num_wavelengths:
            msg = f"Unable to find expected number of wavelength values, expected {num_wavelengths}, found {len(wavelengths)}."
            raise AllotropeConversionError(msg)

    @classmethod
    def get_num_wavelengths(cls, num_wavelengths_raw: str | None) -> int:
        return try_int_or_none(num_wavelengths_raw) or 1

    @classmethod
    def get_wavelengths(cls, wavelengths_str: str | None) -> list[float]:
        return [
            try_float(wavelength, "wavelength")
            for wavelength in assert_not_none(
                wavelengths_str,
                msg="Unable to find wavelengths list.",
            ).split()
        ]

    def iter_wells(self) -> Iterator[str]:
        cols, rows = NUM_WELLS_TO_PLATE_DIMENSIONS[self.header.num_wells]
        for row in range(rows):
            for col in range(1, cols + 1):
                yield f"{num_to_chars(row)}{col}"

    def iter_data_elements(self, position: str) -> Iterator[DataElement]:
        yield from self.block_data.iter_data_elements(position)

    def iter_reduced_data(self) -> Iterator[ReducedDataElement]:
        if self.block_data.reduced_data:
            yield from self.block_data.reduced_data.data


@dataclass(frozen=True)
class FluorescencePlateBlock(PlateBlock):
    @staticmethod
    def get_plate_block_type() -> str:
        return "Fluorescence"

    @classmethod
    def parse_header(cls, header: pd.Series[str]) -> PlateHeader:
        [
            _,  # Plate:
            name,
            export_version,
            export_format,
            read_type,
            _,  # Read mode
            raw_scan_position,
            data_type,
            _,  # Pre-read, always FALSE
            kinetic_points_raw,
            _,  # read_time_or_scan_pattern
            _,  # read_interval_or_scan_density
            _,  # start_wavelength
            _,  # end_wavelength
            _,  # wavelength_step
            num_wavelengths_raw,
            wavelengths_str,
            _,  # first_column
            _,  # num_columns
            num_wells_raw,
            excitation_wavelengths_str,
            _,  # cutoff
            cutoff_filters_str,
            _,  # sweep_wave
            _,  # sweep_wavelength
            reads_per_well,
            pmt_gain,
            _,  # start_integration_time
            _,  # end_integration_time
            _,  # first_row
            num_rows,
        ] = header[:31]

        cls.check_export_version(export_version)
        cls.check_read_type(read_type)
        cls.check_data_type(data_type)

        num_wavelengths = cls.get_num_wavelengths(num_wavelengths_raw)
        wavelengths = cls.get_wavelengths(wavelengths_str)
        cls.check_num_wavelengths(wavelengths, num_wavelengths)

        assert_not_none(
            excitation_wavelengths_str,
            msg="Unable to find excitation wavelengths.",
        )

        excitation_wavelengths = [
            try_int(excitation_wavelength, "excitation wavelengths")
            for excitation_wavelength in excitation_wavelengths_str.split()
        ]

        if len(excitation_wavelengths) != num_wavelengths:
            msg = f"Unable to find expected number of excitation values, expected {num_wavelengths}, found {len(excitation_wavelengths)}"
            raise AllotropeConversionError(msg)

        # cutoff filters is an optional field in the input file
        # if present it contains a list of string numbers separated by spaces
        cutoff_filters = (
            [
                try_int(cutoff_filters, "cutoff filters")
                for cutoff_filters in cutoff_filters_str.split()
            ]
            if cutoff_filters_str is not None
            else None
        )

        # if there are cutoff filters check that the size match number of wavelengths
        if cutoff_filters is not None and len(cutoff_filters) != num_wavelengths:
            msg = f"Unable to find expected number of cutoff filter values, expected {num_wavelengths}, found {len(cutoff_filters)}."
            raise AllotropeConversionError(msg)

        if raw_scan_position == "TRUE":
            scan_position = ScanPosition.BOTTOM
        elif raw_scan_position == "FALSE":
            scan_position = ScanPosition.TOP
        else:
            msg = f"{raw_scan_position} is not a valid scan position, expected 'TRUE' or 'FALSE'."
            raise AllotropeConversionError(msg)

        num_wells = try_int(num_wells_raw, "num_wells")
        num_columns = num_wells_to_n_columns(num_wells)

        return PlateHeader(
            name=name,
            export_version=export_version,
            export_format=export_format,
            read_type=read_type,
            data_type=data_type,
            kinetic_points=try_int(kinetic_points_raw, "kinetic_points"),
            num_wavelengths=num_wavelengths,
            wavelengths=wavelengths,
            num_columns=num_columns,
            num_wells=num_wells,
            concept="fluorescence",
            read_mode="Fluorescence",
            unit="RFU",
            scan_position=scan_position,
            reads_per_well=try_float(reads_per_well, "reads_per_well"),
            pmt_gain=pmt_gain,
            num_rows=try_int(num_rows, "num_rows"),
            excitation_wavelengths=excitation_wavelengths,
            cutoff_filters=cutoff_filters,
        )


@dataclass(frozen=True)
class LuminescencePlateBlock(PlateBlock):
    @staticmethod
    def get_plate_block_type() -> str:
        return "Luminescence"

    @classmethod
    def parse_header(cls, header: pd.Series[str]) -> PlateHeader:
        [
            _,  # Plate:
            name,
            export_version,
            export_format,
            read_type,
            _,  # Read mode
            data_type,
            _,  # Pre-read, always FALSE
            kinetic_points_raw,
            _,  # read_time_or_scan_pattern
            _,  # read_interval_or_scan_density
            _,  # start_wavelength
            _,  # end_wavelength
            _,  # wavelength_step
            num_wavelengths_raw,
            wavelengths_str,
            _,  # first_column
            _,  # num_columns,
            num_wells_raw,
            _,  # excitation_wavelengths_str
            _,  # cutoff
            _,  # cutoff_filters_str,
            _,  # sweep_wave
            _,  # sweep_wavelength
            reads_per_well,
            pmt_gain,
            _,  # start_integration_time
            _,  # end_integration_time
            _,  # first_row
            num_rows,
        ] = header[:30]

        cls.check_export_version(export_version)
        cls.check_read_type(read_type)
        cls.check_data_type(data_type)

        num_wavelengths = cls.get_num_wavelengths(num_wavelengths_raw)
        wavelengths = cls.get_wavelengths(wavelengths_str)
        cls.check_num_wavelengths(wavelengths, num_wavelengths)

        num_wells = try_int(num_wells_raw, "num_wells")
        num_columns = num_wells_to_n_columns(num_wells)

        return PlateHeader(
            name=name,
            export_version=export_version,
            export_format=export_format,
            read_type=read_type,
            data_type=data_type,
            kinetic_points=try_int(kinetic_points_raw, "kinetic_points"),
            num_wavelengths=num_wavelengths,
            wavelengths=wavelengths,
            num_columns=num_columns,
            num_wells=num_wells,
            concept="luminescence",
            read_mode="Luminescence",
            unit="RLU",
            scan_position=ScanPosition.NONE,
            reads_per_well=try_int(reads_per_well, "reads_per_well"),
            pmt_gain=pmt_gain,
            num_rows=try_int(num_rows, "num_rows"),
            excitation_wavelengths=None,
            cutoff_filters=None,
        )


@dataclass(frozen=True)
class AbsorbancePlateBlock(PlateBlock):
    @staticmethod
    def get_plate_block_type() -> str:
        return "Absorbance"

    @classmethod
    def parse_header(cls, header: pd.Series[str]) -> PlateHeader:
        [
            _,  # Plate:
            name,
            export_version,
            export_format,
            read_type,
            _,  # Read mode
            data_type,
            _,  # Pre-read, always FALSE
            kinetic_points_raw,
            _,  # read_time_or_scan_pattern
            _,  # read_interval_or_scan_density
            _,  # start_wavelength
            _,  # end_wavelength
            _,  # wavelength_step
            num_wavelengths_raw,
            wavelengths_str,
            _,  # first_column
            _,  # num_columns
            num_wells_raw,
            _,
            num_rows_raw,
        ] = header[:21]

        cls.check_export_version(export_version)
        cls.check_read_type(read_type)
        cls.check_data_type(data_type)

        num_wavelengths = cls.get_num_wavelengths(num_wavelengths_raw)
        wavelengths = cls.get_wavelengths(wavelengths_str)
        cls.check_num_wavelengths(wavelengths, num_wavelengths)

        num_wells = try_int(num_wells_raw, "num_wells")
        num_columns = num_wells_to_n_columns(num_wells)

        return PlateHeader(
            name=name,
            export_version=export_version,
            export_format=export_format,
            read_type=read_type,
            data_type=data_type,
            kinetic_points=try_int(kinetic_points_raw, "kinetic_points"),
            num_wavelengths=num_wavelengths,
            wavelengths=wavelengths,
            num_columns=num_columns,
            num_wells=num_wells,
            concept="absorbance",
            read_mode="Absorbance",
            unit="mAU",
            scan_position=ScanPosition.NONE,
            reads_per_well=0,
            pmt_gain=None,
            num_rows=try_int(num_rows_raw, "num_rows"),
            excitation_wavelengths=None,
            cutoff_filters=None,
        )


@dataclass(frozen=True)
class BlockList:
    plate_blocks: dict[str, PlateBlock]
    group_blocks: list[GroupBlock]

    @staticmethod
    def create(reader: CsvReader) -> BlockList:
        plate_blocks = {}
        group_blocks = []

        for sub_reader in BlockList._iter_blocks_reader(reader):
            if sub_reader.match("^Group"):
                if "WellPlateName" in assert_not_none(
                    sub_reader.get_line(sub_reader.current_line + 1),
                    msg="Unable to get columns from group block",
                ):
                    group_blocks.append(GroupBlock.create(sub_reader))
            elif sub_reader.match("^Plate"):
                header_series = PlateBlock.read_header(sub_reader)
                cls = PlateBlock.get_plate_block_cls(header_series)
                header = cls.parse_header(header_series)

                export_format_to_data_format = {
                    ExportFormat.TIME_FORMAT.value: TimeData,
                    ExportFormat.PLATE_FORMAT.value: PlateData,
                }
                data_format: type[TimeData] | type[PlateData] = get_key_or_error(
                    "export format", header.export_format, export_format_to_data_format
                )
                block_data = data_format.create(sub_reader, header)

                plate_blocks[header.name] = cls(
                    block_type="Plate",
                    header=header,
                    block_data=block_data,
                )
            elif not sub_reader.match("^Note"):
                msg = f"Expected block '{sub_reader.get()}' to start with Group, Plate or Note."
                raise AllotropeConversionError(msg)

        return BlockList(
            plate_blocks=plate_blocks,
            group_blocks=group_blocks,
        )

    @staticmethod
    def _get_n_blocks(reader: CsvReader) -> int:
        start_line = reader.pop() or ""
        if search_result := re.search(BLOCKS_LINE_REGEX, start_line):
            return int(search_result.group(1))
        msg = f"Unrecognized start line, expected a line starting with ##BLOCKS, got {start_line}"
        raise AllotropeConversionError(msg)

    @staticmethod
    def _iter_blocks_reader(reader: CsvReader) -> Iterator[CsvReader]:
        n_blocks = BlockList._get_n_blocks(reader)
        for _ in range(n_blocks):
            yield CsvReader(list(reader.pop_until(END_LINE_REGEX)))
            reader.pop()  # drop end line
            reader.drop_empty()


@dataclass(frozen=True)
class Data:
    block_list: BlockList

    @staticmethod
    def create(reader: CsvReader) -> Data:
        block_list = BlockList.create(reader)

        for group_block in block_list.group_blocks:
            for group_sample_data in group_block.group_data.sample_data:
                for group_data_element in group_sample_data.data_elements:
                    plate_block = block_list.plate_blocks[group_data_element.plate]
                    for data_element in plate_block.iter_data_elements(
                        group_data_element.position
                    ):
                        data_element.sample_id = group_data_element.sample

        return Data(block_list)



""" Constants file for ThermoFisher Qubit 4 Adapter"""

# Instrument Software Details
QUBIT_SOFTWARE = "Qubit 4 software"
MODEL_NUMBER = "Qubit 4"
BRAND_NAME = "Qubit"
PRODUCT_MANUFACTURER = "Thermo Fisher Scientific"
DEVICE_TYPE = "fluorescence detector"
DISPLAY_NAME = "Thermo Fisher Qubit 4"

# Errors
VALUE_ERROR = "Unable to find value for column"
UNSUPPORTED_FILE_FORMAT_ERROR = "Unsupported file format. Expected xlsx. Actual: "



from __future__ import annotations

from typing import Any

import numpy as np
import pandas as pd

from allotropy.allotrope.pandas_util import read_csv
from allotropy.parsers.roche_cedex_bioht.constants import (
    ANALYTES_LOOKUP,
    DATA_HEADER,
    INFO_HEADER,
    SAMPLE_ROLE_TYPES,
)
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.types import IOType


def to_num(data: pd.Series[Any]) -> pd.Series[Any]:
    return pd.to_numeric(data, errors="coerce").replace(np.nan, None)


class RocheCedexBiohtReader:
    def __init__(self, contents: IOType):
        self.title_data = self.read_title_data(contents)
        self.samples_data = self.read_samples_data(contents)

    def read_title_data(self, contents: IOType) -> SeriesData:
        contents.seek(0)
        return SeriesData(
            read_csv(
                contents,
                delimiter="\t",
                usecols=INFO_HEADER,
                names=INFO_HEADER,
                nrows=1,
            ).T[0]
        )

    def read_samples_data(self, contents: IOType) -> pd.DataFrame:
        contents.seek(0)
        sample_rows = read_csv(
            contents,
            delimiter="\t",
            usecols=DATA_HEADER,
            names=DATA_HEADER,
            skiprows=[0],
        )

        sample_rows = sample_rows.replace(
            {"analyte name": ANALYTES_LOOKUP, "sample role type": SAMPLE_ROLE_TYPES}
        )

        sample_rows["batch identifier"] = sample_rows["batch identifier"].fillna("")

        # concentration values under the test threshold will have a preceding < character
        # this will turn those values into NaN, which is the expected output
        sample_rows["concentration value"] = to_num(sample_rows["concentration value"])

        return sample_rows



from collections.abc import Iterator

from allotropy.parsers.appbio_quantstudio.decorators import cache
from allotropy.parsers.appbio_quantstudio.views import ViewData
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_structure import (
    WellItem,
)
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
    DataSource,
)
from allotropy.parsers.utils.uuids import random_uuid_str


@cache
def build_quantity(well_item: WellItem) -> CalculatedDocument | None:
    if (quantity := well_item.result.quantity) is None:
        return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="quantity",
        value=quantity,
        data_sources=[
            DataSource(feature="cycle threshold result", reference=well_item),
            DataSource(feature="Y-intercept", reference=well_item),
            DataSource(feature="Slope", reference=well_item),
        ],
    )


@cache
def build_quantity_mean(
    view_data: ViewData[WellItem], sample: str, target: str
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(sample, target)
    if (quantity_mean := well_items[0].result.quantity_mean) is None:
        return None

    data_sources = []
    for well_item in well_items:
        quantity_ref = build_quantity(well_item)
        if quantity_ref is None:
            return None

        data_sources.append(
            DataSource(
                feature="quantity",
                reference=quantity_ref,
            )
        )

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="quantity mean",
        value=quantity_mean,
        data_sources=data_sources,
    )


@cache
def build_quantity_sd(
    view_data: ViewData[WellItem], sample: str, target: str
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(sample, target)
    if (quantity_sd := well_items[0].result.quantity_sd) is None:
        return None

    data_sources = []
    for well_item in well_items:
        quantity_ref = build_quantity(well_item)
        if quantity_ref is None:
            return None

        data_sources.append(
            DataSource(
                feature="quantity",
                reference=quantity_ref,
            )
        )

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="quantity sd",
        value=quantity_sd,
        data_sources=data_sources,
    )


@cache
def build_ct_mean(
    view_data: ViewData[WellItem], sample: str, target: str
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(sample, target)
    if (ct_mean := well_items[0].result.ct_mean) is None:
        return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="ct mean",
        value=ct_mean,
        data_sources=[
            DataSource(feature="cycle threshold result", reference=well_item)
            for well_item in well_items
        ],
    )


@cache
def build_ct_sd(
    view_data: ViewData[WellItem], sample: str, target: str
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(sample, target)
    if (ct_sd := well_items[0].result.ct_sd) is None:
        return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="ct sd",
        value=ct_sd,
        data_sources=[
            DataSource(feature="cycle threshold result", reference=well_item)
            for well_item in well_items
        ],
    )


@cache
def build_ct_se(
    view_data: ViewData[WellItem], sample: str, target: str
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(sample, target)
    if (ct_se := well_items[0].result.ct_se) is None:
        return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="ct se",
        value=ct_se,
        data_sources=[
            DataSource(feature="cycle threshold result", reference=well_item)
            for well_item in well_items
        ],
    )


@cache
def build_eq_ct_mean(
    view_data: ViewData[WellItem], sample: str, target: str
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(sample, target)
    if (eq_ct_mean := well_items[0].result.eq_ct_mean) is None:
        return None

    ct_mean_ref = build_ct_mean(view_data, sample, target)
    if ct_mean_ref is None:
        return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="equivalent ct mean",
        value=eq_ct_mean,
        data_sources=[DataSource(feature="ct mean", reference=ct_mean_ref)],
    )


@cache
def build_adj_eq_ct_mean(
    view_data: ViewData[WellItem], sample: str, target: str
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(sample, target)
    if (adj_eq_ct_mean := well_items[0].result.adj_eq_ct_mean) is None:
        return None

    eq_ct_mean_ref = build_eq_ct_mean(view_data, sample, target)
    if eq_ct_mean_ref is None:
        return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="adjusted equivalent ct mean",
        value=adj_eq_ct_mean,
        data_sources=[
            DataSource(feature="equivalent ct mean", reference=eq_ct_mean_ref)
        ],
    )


@cache
def build_delta_ct_mean(
    view_data: ViewData[WellItem],
    sample: str,
    target: str,
    r_target: str | None,
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(sample, target)
    if (delta_ct_mean := well_items[0].result.delta_ct_mean) is None:
        return None

    data_sources = []

    if adj_eq_ct_mean_ref := build_adj_eq_ct_mean(view_data, sample, target):
        data_sources.append(
            DataSource(
                feature="adjusted equivalent ct mean",
                reference=adj_eq_ct_mean_ref,
            )
        )
    elif eq_ct_mean_ref := build_eq_ct_mean(view_data, sample, target):
        data_sources.append(
            DataSource(
                feature="equivalent ct mean",
                reference=eq_ct_mean_ref,
            )
        )
    else:
        return None

    if r_target is not None:
        if r_adj_eq_ct_mean_ref := build_adj_eq_ct_mean(view_data, sample, r_target):
            data_sources.append(
                DataSource(
                    feature="adjusted equivalent ct mean",
                    reference=r_adj_eq_ct_mean_ref,
                )
            )
        elif r_eq_ct_mean_ref := build_eq_ct_mean(view_data, sample, r_target):
            data_sources.append(
                DataSource(
                    feature="equivalent ct mean",
                    reference=r_eq_ct_mean_ref,
                )
            )
        else:
            return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="delta equivalent ct mean",
        value=delta_ct_mean,
        data_sources=data_sources,
    )


@cache
def build_delta_ct_sd(
    view_data: ViewData[WellItem], sample: str, target: str, r_target: str | None
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(sample, target)
    if (delta_ct_sd := well_items[0].result.delta_ct_sd) is None:
        return None

    data_sources = []

    if ct_sd_ref := build_ct_sd(view_data, sample, target):
        data_sources.append(DataSource(feature="ct sd", reference=ct_sd_ref))
    else:
        return None

    if r_target is not None:
        if r_ct_sd_ref := build_ct_sd(view_data, sample, r_target):
            data_sources.append(DataSource(feature="ct sd", reference=r_ct_sd_ref))
        else:
            return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="delta equivalent ct sd",
        value=delta_ct_sd,
        data_sources=data_sources,
    )


@cache
def build_delta_ct_se(
    view_data: ViewData[WellItem], sample: str, target: str, r_target: str | None
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(sample, target)
    if (delta_ct_se := well_items[0].result.delta_ct_se) is None:
        return None

    data_sources = []
    if ct_se_ref := build_ct_se(view_data, sample, target):
        data_sources.append(DataSource(feature="ct se", reference=ct_se_ref))
    else:
        return None

    if r_target is not None:
        if r_ct_se_ref := build_ct_se(view_data, sample, r_target):
            data_sources.append(DataSource(feature="ct se", reference=r_ct_se_ref))
        else:
            return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="delta equivalent ct se",
        value=delta_ct_se,
        data_sources=data_sources,
    )


@cache
def build_delta_delta_ct(
    view_data: ViewData[WellItem],
    sample: str,
    target: str,
    r_sample: str,
    r_target: str | None,
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(sample, target)
    if (delta_delta_ct := well_items[0].result.delta_delta_ct) is None:
        return None

    delta_ct_mean_ref = build_delta_ct_mean(view_data, sample, target, r_target)
    if delta_ct_mean_ref is None:
        return None

    r_delta_ct_mean_ref = build_delta_ct_mean(view_data, r_sample, target, r_target)
    if r_delta_ct_mean_ref is None:
        return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="delta delta equivalent ct",
        value=delta_delta_ct,
        data_sources=[
            DataSource(
                feature="delta equivalent ct mean",
                reference=delta_ct_mean_ref,
            ),
            DataSource(
                feature="delta equivalent ct mean",
                reference=r_delta_ct_mean_ref,
            ),
        ],
    )


@cache
def build_rq(
    view_data: ViewData[WellItem],
    sample: str,
    target: str,
    r_sample: str,
    r_target: str | None,
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(sample, target)
    if (rq := well_items[0].result.rq) is None:
        return None

    delta_delta_ct_ref = build_delta_delta_ct(
        view_data, sample, target, r_sample, r_target
    )
    if delta_delta_ct_ref is None:
        return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="rq",
        value=rq,
        data_sources=[
            DataSource(
                feature="delta delta equivalent ct",
                reference=delta_delta_ct_ref,
            )
        ],
    )


@cache
def build_rq_min(
    view_data: ViewData[WellItem],
    sample: str,
    target: str,
    r_sample: str,
    r_target: str | None,
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(sample, target)
    if (rq_min := well_items[0].result.rq_min) is None:
        return None

    rq_ref = build_rq(view_data, sample, target, r_sample, r_target)
    if rq_ref is None:
        return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="rq min",
        value=rq_min,
        data_sources=[
            DataSource(
                feature="rq",
                reference=rq_ref,
            )
        ],
    )


@cache
def build_rq_max(
    view_data: ViewData[WellItem],
    sample: str,
    target: str,
    r_sample: str,
    r_target: str,
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(sample, target)
    if (rq_max := well_items[0].result.rq_max) is None:
        return None

    rq_ref = build_rq(view_data, sample, target, r_sample, r_target)
    if rq_ref is None:
        return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="rq max",
        value=rq_max,
        data_sources=[
            DataSource(
                feature="rq",
                reference=rq_ref,
            ),
        ],
    )


@cache
def build_relative_rq(
    view_data: ViewData[WellItem],
    sample: str,
    target: str,
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(sample, target)
    if (rq := well_items[0].result.rq) is None:
        return None

    quantity_mean_ref = build_quantity_mean(view_data, sample, target)
    if quantity_mean_ref is None:
        return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="rq",
        value=rq,
        data_sources=[
            DataSource(
                feature="quantity mean",
                reference=quantity_mean_ref,
            )
        ],
    )


@cache
def build_relative_rq_min(
    view_data: ViewData[WellItem],
    sample: str,
    target: str,
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(sample, target)
    if (rq_min := well_items[0].result.rq_min) is None:
        return None

    relative_rq_ref = build_relative_rq(view_data, sample, target)
    if relative_rq_ref is None:
        return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="rq min",
        value=rq_min,
        data_sources=[
            DataSource(
                feature="rq",
                reference=relative_rq_ref,
            )
        ],
    )


@cache
def build_relative_rq_max(
    view_data: ViewData[WellItem],
    sample: str,
    target: str,
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(sample, target)
    if (rq_max := well_items[0].result.rq_max) is None:
        return None

    relative_rq_ref = build_relative_rq(view_data, sample, target)
    if relative_rq_ref is None:
        return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="rq max",
        value=rq_max,
        data_sources=[
            DataSource(
                feature="rq",
                reference=relative_rq_ref,
            ),
        ],
    )


@cache
def build_rn_mean(
    view_data: ViewData[WellItem], sample: str, target: str
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(sample, target)
    if (rn_mean := well_items[0].result.rn_mean) is None:
        return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="rn mean",
        value=rn_mean,
        data_sources=[
            DataSource(feature="normalized reporter result", reference=well_item)
            for well_item in well_items
        ],
    )


@cache
def build_rn_sd(
    view_data: ViewData[WellItem], sample: str, target: str
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(sample, target)
    if (rn_sd := well_items[0].result.rn_sd) is None:
        return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="rn sd",
        value=rn_sd,
        data_sources=[
            DataSource(feature="normalized reporter result", reference=well_item)
            for well_item in well_items
        ],
    )


@cache
def build_y_intercept(
    view_data: ViewData[WellItem], target: str
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(target)
    if (y_intercept := well_items[0].result.y_intercept) is None:
        return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="y intercept",
        value=y_intercept,
        data_sources=[
            DataSource(feature="cycle threshold result", reference=well_item)
            for well_item in well_items
        ],
    )


@cache
def build_r_squared(
    view_data: ViewData[WellItem], target: str
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(target)
    if (r_squared := well_items[0].result.r_squared) is None:
        return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="r^2",
        value=r_squared,
        data_sources=[
            DataSource(feature="cycle threshold result", reference=well_item)
            for well_item in well_items
        ],
    )


@cache
def build_slope(
    view_data: ViewData[WellItem], target: str
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(target)
    if (slope := well_items[0].result.slope) is None:
        return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="slope",
        value=slope,
        data_sources=[
            DataSource(feature="cycle threshold result", reference=well_item)
            for well_item in well_items
        ],
    )


@cache
def build_efficiency(
    view_data: ViewData[WellItem], target: str
) -> CalculatedDocument | None:
    well_items = view_data.get_leaf_item(target)
    if (efficiency := well_items[0].result.efficiency) is None:
        return None

    return CalculatedDocument(
        uuid=random_uuid_str(),
        name="efficiency",
        value=efficiency,
        data_sources=[
            DataSource(feature="cycle threshold result", reference=well_item)
            for well_item in well_items
        ],
    )


def iter_standard_curve_calc_docs(
    view_st_data: ViewData[WellItem],
    view_tr_data: ViewData[WellItem],
) -> Iterator[CalculatedDocument]:
    # Quantity, Quantity Mean, Quantity SD, Ct Mean, Ct SD, Y-Intercept,
    # R(superscript 2), Slope, Efficiency
    for sample, target in view_st_data.iter_keys():
        if calc_doc := build_quantity_mean(view_st_data, sample, target):
            yield from calc_doc.iter_struct()

        if calc_doc := build_quantity_sd(view_st_data, sample, target):
            yield from calc_doc.iter_struct()

        if calc_doc := build_ct_mean(view_st_data, sample, target):
            yield from calc_doc.iter_struct()

        if calc_doc := build_ct_sd(view_st_data, sample, target):
            yield from calc_doc.iter_struct()

    for target in view_tr_data.data:
        if calc_doc := build_y_intercept(view_tr_data, target):
            yield from calc_doc.iter_struct()

        if calc_doc := build_r_squared(view_tr_data, target):
            yield from calc_doc.iter_struct()

        if calc_doc := build_slope(view_tr_data, target):
            yield from calc_doc.iter_struct()

        if calc_doc := build_efficiency(view_tr_data, target):
            yield from calc_doc.iter_struct()


def iter_relative_standard_curve_calc_docs(
    view_data: ViewData[WellItem],
    r_sample: str,
    r_target: str | None,
) -> Iterator[CalculatedDocument]:
    for sample, target in view_data.iter_keys():
        if calc_doc := build_ct_mean(view_data, sample, target):
            yield from calc_doc.iter_struct()

        if calc_doc := build_ct_sd(view_data, sample, target):
            yield from calc_doc.iter_struct()

        if calc_doc := build_delta_ct_sd(view_data, sample, target, r_target):
            yield from calc_doc.iter_struct()

        if calc_doc := build_delta_ct_se(view_data, sample, target, r_target):
            yield from calc_doc.iter_struct()

        if calc_doc := build_relative_rq_min(view_data, sample, target):
            yield from calc_doc.iter_struct()

        if calc_doc := build_relative_rq_max(view_data, sample, target):
            yield from calc_doc.iter_struct()

        if target != r_target:
            if calc_doc := build_rq_min(view_data, sample, target, r_sample, r_target):
                yield from calc_doc.iter_struct()

            if calc_doc := build_rq_max(view_data, sample, target, r_sample, r_target):
                yield from calc_doc.iter_struct()


def iter_presence_absence_calc_docs(
    view_data: ViewData[WellItem],
) -> Iterator[CalculatedDocument]:
    # Rn Mean, Rn SD
    for sample, target in view_data.iter_keys():
        if calc_doc := build_rn_mean(view_data, sample, target):
            yield from calc_doc.iter_struct()

        if calc_doc := build_rn_sd(view_data, sample, target):
            yield from calc_doc.iter_struct()



from __future__ import annotations

import pandas as pd

from allotropy.allotrope.models.shared.definitions.definitions import NaN
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    CalculatedDataItem,
    Data,
    DataSource,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.unchained_labs_lunatic.constants import (
    CALCULATED_DATA_LOOKUP,
    INCORRECT_WAVELENGTH_COLUMN_FORMAT_ERROR_MSG,
    NO_DATE_OR_TIME_ERROR_MSG,
    NO_MEASUREMENT_IN_PLATE_ERROR_MSG,
    NO_WAVELENGTH_COLUMN_ERROR_MSG,
    WAVELENGTH_COLUMNS_RE,
)
from allotropy.parsers.utils.pandas import map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


def _create_measurement(
    well_plate_data: SeriesData,
    wavelength_column: str,
    calculated_data: list[CalculatedDataItem],
) -> Measurement:
    if wavelength_column not in well_plate_data.series:
        msg = NO_MEASUREMENT_IN_PLATE_ERROR_MSG.format(wavelength_column)
        raise AllotropeConversionError(msg)

    if not WAVELENGTH_COLUMNS_RE.match(wavelength_column):
        raise AllotropeConversionError(INCORRECT_WAVELENGTH_COLUMN_FORMAT_ERROR_MSG)

    measurement_identifier = random_uuid_str()
    calculated_data.extend(
        _get_calculated_data(well_plate_data, wavelength_column, measurement_identifier)
    )
    return Measurement(
        type_=MeasurementType.ULTRAVIOLET_ABSORBANCE,
        identifier=measurement_identifier,
        detector_wavelength_setting=float(wavelength_column[1:]),
        absorbance=well_plate_data.get(float, wavelength_column, NaN),
        sample_identifier=well_plate_data[str, "Sample name"],
        location_identifier=well_plate_data[str, "Plate Position"],
        well_plate_identifier=well_plate_data.get(str, "Plate ID"),
    )


def _get_calculated_data(
    well_plate_data: SeriesData,
    wavelength_column: str,
    measurement_identifier: str,
) -> list[CalculatedDataItem]:
    calculated_data = []
    for item in CALCULATED_DATA_LOOKUP.get(wavelength_column, []):
        value = well_plate_data.get(float, item["column"])
        if value is None:
            continue

        calculated_data.append(
            CalculatedDataItem(
                identifier=random_uuid_str(),
                name=item["name"],
                value=value,
                unit=item["unit"],
                data_sources=[
                    DataSource(
                        identifier=measurement_identifier,
                        feature=item["feature"],
                    )
                ],
            )
        )
    return calculated_data


def _create_measurement_group(
    data: SeriesData,
    wavelength_columns: list[str],
    calculated_data: list[CalculatedDataItem],
) -> MeasurementGroup:
    date = data.get(str, "Date")
    time = data.get(str, "Time")

    if not date or not time:
        raise AllotropeConversionError(NO_DATE_OR_TIME_ERROR_MSG)

    return MeasurementGroup(
        _measurement_time=f"{date} {time}",
        analytical_method_identifier=data.get(str, "Application"),
        plate_well_count=96,
        measurements=[
            _create_measurement(data, wavelength_column, calculated_data)
            for wavelength_column in wavelength_columns
        ],
    )


def _create_metadata(data: pd.DataFrame, file_name: str) -> Metadata:
    return Metadata(
        device_type="plate reader",
        model_number="Lunatic",
        product_manufacturer="Unchained Labs",
        device_identifier=SeriesData(data.iloc[0])[str, "Instrument ID"],
        software_name="Lunatic and Stunner Analysis",
        file_name=file_name,
    )


def create_data(data: pd.DataFrame, file_name: str) -> Data:
    wavelength_columns = list(filter(WAVELENGTH_COLUMNS_RE.match, data.columns))
    if not wavelength_columns:
        raise AllotropeConversionError(NO_WAVELENGTH_COLUMN_ERROR_MSG)

    # TODO: we are reporting calculated data for measurements globally instead of in the measurement doc,
    # which is why we have to pass this list to collect them. Why are we reporting globally when data is
    # pertains to the individual measurements?
    calculated_data: list[CalculatedDataItem] = []

    def make_group(data: SeriesData) -> MeasurementGroup:
        return _create_measurement_group(data, wavelength_columns, calculated_data)

    return Data(
        metadata=_create_metadata(data, file_name),
        measurement_groups=map_rows(data, make_group),
        calculated_data=calculated_data,
    )



from __future__ import annotations

from allotropy.allotrope.schema_mappers.adm.cell_counting.benchling._2023._11.cell_counting import (
    Data,
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.exceptions import (
    AllotropeConversionError,
)
from allotropy.parsers.beckman_vi_cell_xr.constants import (
    DEFAULT_ANALYST,
    MODEL_NUMBER,
    SOFTWARE_NAME,
)
from allotropy.parsers.beckman_vi_cell_xr.vi_cell_xr_reader import ViCellData
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


def _create_measurement_group(data: SeriesData) -> MeasurementGroup:
    total_cell_count = data.get(float, "Total cells")
    total_cell_count = (
        total_cell_count if total_cell_count is None else round(total_cell_count)
    )
    viable_cell_count = data.get(float, "Viable cells")
    viable_cell_count = (
        viable_cell_count if viable_cell_count is None else round(viable_cell_count)
    )

    return MeasurementGroup(
        measurements=[
            Measurement(
                measurement_identifier=random_uuid_str(),
                timestamp=data[str, "Sample date"],
                sample_identifier=data[str, "Sample ID"],
                cell_type_processing_method=data.get(str, "Cell type"),
                cell_density_dilution_factor=data.get(float, "Dilution factor"),
                viability=data[float, "Viability (%)"],
                viable_cell_density=data[float, "Viable cells/ml (x10^6)"],
                total_cell_count=total_cell_count,
                total_cell_density=data.get(float, "Total cells/ml (x10^6)"),
                average_total_cell_diameter=data.get(float, "Avg. diam. (microns)"),
                viable_cell_count=viable_cell_count,
                average_total_cell_circularity=data.get(float, "Avg. circ."),
                analyst=DEFAULT_ANALYST,
            )
        ]
    )


def create_data(reader_data: ViCellData, file_name: str) -> Data:
    if not reader_data.data:
        msg = "Cannot parse ASM from empty file."
        raise AllotropeConversionError(msg)

    metadata = Metadata(
        device_type="brightfield imager (cell counter)",
        detection_type="brightfield",
        model_number=MODEL_NUMBER,
        equipment_serial_number=reader_data.serial_number,
        software_name=SOFTWARE_NAME,
        software_version=reader_data.version.value,
        file_name=file_name,
    )

    return Data(metadata, [_create_measurement_group(row) for row in reader_data.data])



from __future__ import annotations

from collections import defaultdict
from typing import Any, cast, TypeVar

import pandas as pd

from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    ContainerType,
    DataSystemDocument,
    DeviceControlDocument,
    DeviceSystemDocument,
    FluorescencePointDetectionDeviceControlAggregateDocument,
    FluorescencePointDetectionDeviceControlDocumentItem,
    FluorescencePointDetectionMeasurementDocumentItems,
    LuminescencePointDetectionMeasurementDocumentItems,
    MeasurementAggregateDocument,
    Model,
    OpticalImagingMeasurementDocumentItems,
    PlateReaderAggregateDocument,
    PlateReaderDocumentItem,
    SampleDocument,
    UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument,
    UltravioletAbsorbancePointDetectionDeviceControlDocumentItem,
    UltravioletAbsorbancePointDetectionMeasurementDocumentItems,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueNanometer,
    TQuantityValueNumber,
    TQuantityValueRelativeFluorescenceUnit,
)
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.exceptions import valid_value_or_raise
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.bmg_mars.bmg_mars_structure import (
    get_plate_data,
    get_plate_well_count,
    Header,
    ReadType,
    Wavelength,
)
from allotropy.parsers.lines_reader import LinesReader, read_to_lines
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser

MeasurementDocumentItems = (
    FluorescencePointDetectionMeasurementDocumentItems
    | UltravioletAbsorbancePointDetectionMeasurementDocumentItems
    | OpticalImagingMeasurementDocumentItems
    | LuminescencePointDetectionMeasurementDocumentItems
)

DeviceControlAggregateDocument = (
    FluorescencePointDetectionDeviceControlAggregateDocument
    | UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument
)

T = TypeVar("T")


def safe_value(cls: type[T], value: Any | None) -> T | None:
    return None if value is None else cls(value=value)  # type: ignore[call-arg]


class BmgMarsParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "BMG MARS"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.WORKING_DRAFT

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        filename = named_file_contents.original_file_name
        lines = read_to_lines(named_file_contents)
        read_type = self._get_read_type(lines)

        reader = LinesReader(lines)
        raw_header = list(reader.pop_until("^,?Raw Data"))
        header = Header.create(raw_header)

        csv_data = list(reader.pop_until_empty())
        wavelength = Wavelength.create(csv_data)
        plate_well_count = get_plate_well_count(csv_data)
        data = get_plate_data(csv_data)

        return self._get_model(
            header,
            data,
            filename,
            read_type,
            wavelength,
            plate_well_count,
        )

    def _get_model(
        self,
        header: Header,
        data: pd.DataFrame,
        filename: str,
        read_type: ReadType,
        wavelength: Wavelength,
        plate_well_count: TQuantityValueNumber,
    ) -> Model:

        return Model(
            plate_reader_aggregate_document=PlateReaderAggregateDocument(
                plate_reader_document=self._get_plate_reader_document(
                    data, read_type, header, wavelength, plate_well_count
                ),
                device_system_document=DeviceSystemDocument(
                    device_identifier="N/A",
                    model_number="N/A",
                    product_manufacturer="BMG LABTECH",
                ),
                data_system_document=DataSystemDocument(
                    file_name=filename,
                    UNC_path=header.path,
                    software_name="BMG MARS",
                    ASM_converter_name=self.get_asm_converter_name(),
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
            ),
            field_asm_manifest="http://purl.allotrope.org/manifests/plate-reader/BENCHLING/2023/09/plate-reader.manifest",
        )

    def _get_read_type(self, lines: list[str]) -> ReadType:
        content = "\n".join(lines).lower()
        read_types = {
            read_type for read_type in ReadType if read_type.value.lower() in content
        }
        return valid_value_or_raise("read type", read_types, ReadType)

    def _get_device_control_aggregate_document(
        self,
        wavelength: Wavelength,
        read_type: ReadType,
    ) -> DeviceControlAggregateDocument:

        if read_type == ReadType.ABSORBANCE:
            return UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument(
                device_control_document=[
                    UltravioletAbsorbancePointDetectionDeviceControlDocumentItem(
                        device_type="absorbance detector",
                        detection_type="absorbance",
                        detector_wavelength_setting=TQuantityValueNanometer(
                            value=wavelength.wavelength
                        ),
                    )
                ]
            )

        elif read_type == ReadType.FLUORESCENCE:
            return FluorescencePointDetectionDeviceControlAggregateDocument(
                device_control_document=[
                    FluorescencePointDetectionDeviceControlDocumentItem(
                        device_type="fluorescence detector",
                        detection_type="fluorescence",
                        detector_wavelength_setting=TQuantityValueNanometer(
                            value=wavelength.wavelength
                        ),
                        excitation_wavelength_setting=TQuantityValueNanometer(
                            value=wavelength.ex_wavelength
                        ),
                    )
                ]
            )

    def _get_measurement_document(
        self,
        read_type: ReadType,
        header: Header,
        result: pd.Series[Any],
        device_control_document: list[DeviceControlDocument],
    ) -> MeasurementDocumentItems:
        plate_barcode = header.id1
        well_location = f"{result.row}{result.col}"
        sample_document = SampleDocument(
            sample_identifier=f"{plate_barcode} {well_location}",
            location_identifier=well_location,
            well_plate_identifier=plate_barcode,
        )
        if read_type == ReadType.ABSORBANCE:
            return UltravioletAbsorbancePointDetectionMeasurementDocumentItems(
                measurement_identifier=result.uuid,
                sample_document=sample_document,
                device_control_aggregate_document=UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument(
                    device_control_document=cast(
                        list[
                            UltravioletAbsorbancePointDetectionDeviceControlDocumentItem
                        ],
                        device_control_document,
                    )
                ),
                absorbance=TQuantityValueMilliAbsorbanceUnit(value=result.value),
            )
        elif read_type == ReadType.FLUORESCENCE:
            return FluorescencePointDetectionMeasurementDocumentItems(
                measurement_identifier=result.uuid,
                sample_document=sample_document,
                device_control_aggregate_document=FluorescencePointDetectionDeviceControlAggregateDocument(
                    device_control_document=cast(
                        list[FluorescencePointDetectionDeviceControlDocumentItem],
                        device_control_document,
                    )
                ),
                fluorescence=TQuantityValueRelativeFluorescenceUnit(value=result.value),
            )

    def _get_plate_reader_document(
        self,
        data: pd.DataFrame,
        read_type: ReadType,
        header: Header,
        wavelength: Wavelength,
        plate_well_count: TQuantityValueNumber,
    ) -> list[PlateReaderDocumentItem]:
        items = []
        measurement_docs_dict = defaultdict(list)
        measurement_time = self._get_date_time(f"{header.date} {header.time}")
        device_control_aggregate_document = self._get_device_control_aggregate_document(
            wavelength, read_type
        )

        for _, result in data.iterrows():
            measurement_docs_dict[(result.row, result.col)].append(
                self._get_measurement_document(
                    read_type,
                    header,
                    result,
                    cast(
                        list[DeviceControlDocument],
                        device_control_aggregate_document.device_control_document,
                    ),
                )
            )

        for well_location in sorted(measurement_docs_dict.keys()):
            items.append(
                PlateReaderDocumentItem(
                    measurement_aggregate_document=MeasurementAggregateDocument(
                        measurement_time=measurement_time,
                        plate_well_count=plate_well_count,
                        measurement_document=measurement_docs_dict[well_location],
                        experiment_type=header.test_name,
                        experimental_data_identifier=header.id2,
                        container_type=ContainerType.well_plate,
                    ),
                    analyst=header.user,
                )
            )

        return items



# header
# sample setup
# raw data
# amplification data
# multicomponent data
# results
# melt curve raw data

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from io import StringIO
import re
from typing import Any, TypeVar

import numpy as np
import pandas as pd

from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import ExperimentType
from allotropy.allotrope.pandas_util import read_csv
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.lines_reader import LinesReader
from allotropy.parsers.utils.calculated_data_documents.definition import Referenceable
from allotropy.parsers.utils.pandas import df_to_series_data, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import assert_not_none, try_int

T = TypeVar("T")


def map_wells(map_func: Callable[..., T], data: pd.DataFrame) -> dict[int, T]:
    return {
        try_int(str(well_id), "well id"): map_func(well_data)
        for well_id, well_data in data.groupby("Well")
    }


@dataclass(frozen=True)
class Header:
    measurement_time: str
    plate_well_count: int | None
    experiment_type: ExperimentType
    device_identifier: str
    model_number: str
    device_serial_number: str
    measurement_method_identifier: str
    pcr_detection_chemistry: str
    passive_reference_dye_setting: str | None
    barcode: str | None
    analyst: str | None
    experimental_data_identifier: str | None

    @staticmethod
    def create(reader: LinesReader) -> Header:
        lines = [line.replace("*", "", 1) for line in reader.pop_until(r"^\[.+\]")]
        csv_stream = StringIO("\n".join(lines))
        raw_data = read_csv(csv_stream, header=None, sep="=", names=["index", "values"])
        series = pd.Series(raw_data["values"].values, index=raw_data["index"]).astype(
            str
        )
        series.index = series.index.str.strip()
        series = series.str.strip().replace("NA", None)

        data = SeriesData(series)

        experiments_type_options = {
            "Standard Curve": ExperimentType.standard_curve_qPCR_experiment,
            "Relative Standard Curve": ExperimentType.relative_standard_curve_qPCR_experiment,
            "Comparative Cт (ΔΔCт)": ExperimentType.comparative_CT_qPCR_experiment,
            "Melt Curve": ExperimentType.melt_curve_qPCR_experiment,
            "Genotyping": ExperimentType.genotyping_qPCR_experiment,
            "Presence/Absence": ExperimentType.presence_absence_qPCR_experiment,
        }

        plate_well_count_search = re.search("(96)|(384)", data[str, "Block Type"])

        return Header(
            measurement_time=data[str, "Experiment Run End Time"],
            plate_well_count=(
                None
                if plate_well_count_search is None
                else int(plate_well_count_search.group())
            ),
            experiment_type=assert_not_none(
                experiments_type_options.get(
                    data[str, "Experiment Type"],
                ),
                msg="Unable to find valid experiment type",
            ),
            device_identifier=(data.get(str, "Instrument Name", NOT_APPLICABLE)),
            model_number=data[str, "Instrument Type"],
            device_serial_number=data.get(
                str, "Instrument Serial Number", NOT_APPLICABLE
            ),
            measurement_method_identifier=data[str, "Quantification Cycle Method"],
            pcr_detection_chemistry=data[str, "Chemistry"],
            passive_reference_dye_setting=data.get(str, "Passive Reference"),
            barcode=data.get(str, "Experiment Barcode"),
            analyst=data.get(str, "Experiment User Name"),
            experimental_data_identifier=data.get(str, "Experiment Name"),
        )


@dataclass
class WellItem(Referenceable):
    identifier: int
    target_dna_description: str
    sample_identifier: str
    reporter_dye_setting: str | None = None
    position: str | None = None
    well_location_identifier: str | None = None
    quencher_dye_setting: str | None = None
    sample_role_type: str | None = None
    _result: Result | None = None

    @property
    def result(self) -> Result:
        return assert_not_none(self._result)

    @staticmethod
    def create_genotyping(data: SeriesData) -> tuple[WellItem, WellItem]:
        identifier = data[int, "Well"]
        snp_name = data[
            str, "SNP Assay Name", f"Unable to find snp name for well {identifier}"
        ]
        allele1 = data[
            str, "Allele1 Name", f"Unable to find allele 1 for well {identifier}"
        ]
        allele2 = data[
            str, "Allele2 Name", f"Unable to find allele 2 for well {identifier}"
        ]
        return (
            WellItem(
                uuid=random_uuid_str(),
                identifier=identifier,
                target_dna_description=f"{snp_name}-{allele1}",
                sample_identifier=data.get(str, "Sample Name", NOT_APPLICABLE),
                reporter_dye_setting=data.get(str, "Allele1 Reporter"),
                position=data.get(str, "Well Position", NOT_APPLICABLE),
                well_location_identifier=data.get(str, "Well Position"),
                quencher_dye_setting=data.get(str, "Quencher"),
                sample_role_type=data.get(str, "Task"),
            ),
            WellItem(
                uuid=random_uuid_str(),
                identifier=identifier,
                target_dna_description=f"{snp_name}-{allele2}",
                sample_identifier=data.get(str, "Sample Name", NOT_APPLICABLE),
                reporter_dye_setting=data.get(str, "Allele2 Reporter"),
                position=data.get(str, "Well Position", NOT_APPLICABLE),
                well_location_identifier=data.get(str, "Well Position"),
                quencher_dye_setting=data.get(str, "Quencher"),
                sample_role_type=data.get(str, "Task"),
            ),
        )

    @staticmethod
    def create_generic(data: SeriesData) -> WellItem:
        identifier = data[int, "Well"]
        return WellItem(
            uuid=random_uuid_str(),
            identifier=identifier,
            target_dna_description=data[
                str,
                "Target Name",
                f"Unable to find target dna description for well {identifier}",
            ],
            sample_identifier=data.get(str, "Sample Name", NOT_APPLICABLE),
            reporter_dye_setting=data.get(str, "Reporter"),
            position=data.get(str, "Well Position", NOT_APPLICABLE),
            well_location_identifier=data.get(str, "Well Position"),
            quencher_dye_setting=data.get(str, "Quencher"),
            sample_role_type=data.get(str, "Task"),
        )


@dataclass
class Well:
    identifier: int
    items: list[WellItem]

    @staticmethod
    def create_genotyping(series: pd.Series[Any]) -> Well:
        return Well(
            identifier=try_int(str(series.name), "well id"),
            items=list(WellItem.create_genotyping(SeriesData(series))),
        )

    @staticmethod
    def create_generic(well_data: pd.DataFrame) -> Well:
        return Well(
            identifier=try_int(str(well_data["Well"].iloc[0]), "well id"),
            items=[
                WellItem.create_generic(SeriesData(item_data))
                for _, item_data in well_data.iterrows()
            ],
        )

    @staticmethod
    def create(reader: LinesReader, experiment_type: ExperimentType) -> list[Well]:
        assert_not_none(
            reader.drop_until(r"^\[Sample Setup\]"),
            msg="Unable to find 'Sample Setup' section in file.",
        )

        reader.pop()  # remove title
        lines = list(reader.pop_until(r"^\[.+\]"))
        csv_stream = StringIO("\n".join(lines))
        data = read_csv(csv_stream, sep="\t").replace(np.nan, None)

        if experiment_type == ExperimentType.genotyping_qPCR_experiment:
            return list(data.apply(Well.create_genotyping, axis="columns"))  # type: ignore[call-overload]
        else:
            return list(
                map_wells(
                    Well.create_generic, data[data["Target Name"].notnull()]
                ).values()
            )


@dataclass(frozen=True)
class RawData:
    lines: list[str]

    @staticmethod
    def create(reader: LinesReader) -> RawData | None:
        if reader.match(r"^\[Raw Data\]"):
            reader.pop()  # remove title
            return RawData(lines=list(reader.pop_until(r"^\[.+\]")))
        return None


@dataclass(frozen=True)
class AmplificationData:
    total_cycle_number_setting: float
    cycle: list[float]
    rn: list[float | None]
    delta_rn: list[float | None]

    @staticmethod
    def create(reader: LinesReader) -> dict[int, dict[str, AmplificationData]]:
        assert_not_none(
            reader.drop_until(r"^\[Amplification Data\]"),
            msg="Unable to find 'Amplification Data' section in file.",
        )

        reader.pop()  # remove title
        lines = list(reader.pop_until(r"^\[.+\]"))
        csv_stream = StringIO("\n".join(lines))
        data = read_csv(csv_stream, sep="\t", thousands=r",")

        def make_data(well_data: pd.DataFrame) -> dict[str, AmplificationData]:
            return {
                str(target_name): AmplificationData(
                    total_cycle_number_setting=float(target_data["Cycle"].max()),
                    cycle=target_data["Cycle"].tolist(),
                    rn=target_data["Rn"].tolist(),
                    delta_rn=target_data["Delta Rn"].tolist(),
                )
                for target_name, target_data in well_data.groupby("Target Name")
            }

        return map_wells(make_data, data)


@dataclass(frozen=True)
class MulticomponentData:
    cycle: list[float]
    columns: dict[str, list[float | None]]

    def get_column(self, name: str) -> list[float | None]:
        return assert_not_none(
            self.columns.get(name),
            msg=f"Unable to obtain '{name}' from multicomponent data.",
        )

    @staticmethod
    def create(reader: LinesReader) -> dict[int, MulticomponentData]:
        if not reader.match(r"^\[Multicomponent Data\]"):
            return {}
        reader.pop()  # remove title
        lines = list(reader.pop_until(r"^\[.+\]"))
        csv_stream = StringIO("\n".join(lines))
        data = read_csv(csv_stream, sep="\t", thousands=r",")

        def make_data(well_data: pd.Series[Any]) -> MulticomponentData:
            return MulticomponentData(
                cycle=well_data["Cycle"].tolist(),
                columns={
                    name: well_data[name].tolist()
                    for name in well_data
                    if name not in ["Well", "Cycle", "Well Position"]
                },
            )

        return map_wells(make_data, data)


@dataclass(frozen=True)
class ResultMetadata:
    reference_dna_description: str | None
    reference_sample_description: str | None

    @staticmethod
    def create(data: SeriesData, experiment_type: ExperimentType) -> ResultMetadata:
        if experiment_type not in [
            ExperimentType.comparative_CT_qPCR_experiment,
            ExperimentType.relative_standard_curve_qPCR_experiment,
        ]:
            return ResultMetadata(None, None)
        return ResultMetadata(
            reference_dna_description=data.get(
                str, "Endogenous Control", NOT_APPLICABLE
            ),
            reference_sample_description=data.get(
                str, "Reference Sample", NOT_APPLICABLE
            ),
        )


@dataclass(frozen=True)
class Result:
    cycle_threshold_value_setting: float
    cycle_threshold_result: float | None
    automatic_cycle_threshold_enabled_setting: bool | None
    automatic_baseline_determination_enabled_setting: bool | None
    normalized_reporter_result: float | None
    baseline_corrected_reporter_result: float | None
    genotyping_determination_result: str | None
    genotyping_determination_method_setting: float | None
    quantity: float | None
    quantity_mean: float | None
    quantity_sd: float | None
    ct_mean: float | None
    ct_sd: float | None
    delta_ct_mean: float | None
    delta_ct_se: float | None
    delta_delta_ct: float | None
    rq: float | None
    rq_min: float | None
    rq_max: float | None
    rn_mean: float | None
    rn_sd: float | None
    y_intercept: float | None
    r_squared: float | None
    slope: float | None
    efficiency: float | None

    @staticmethod
    def create(
        reader: LinesReader, experiment_type: ExperimentType
    ) -> tuple[dict[int, dict[str, Result]], ResultMetadata]:
        assert_not_none(
            reader.drop_until(r"^\[Results\]"),
            msg="Unable to find 'Results' section in file.",
        )

        reader.pop()  # remove title
        data_lines = list(reader.pop_until_empty())
        csv_stream = StringIO("\n".join(data_lines))
        data = read_csv(csv_stream, sep="\t", thousands=r",").replace(np.nan, None)
        result = Result.create_results(data, experiment_type)

        reader.drop_empty()

        if reader.match(r"\[.+\]"):
            return result, ResultMetadata.create(
                SeriesData(pd.Series()), experiment_type
            )

        metadata_lines = list(reader.pop_until_empty())
        csv_stream = StringIO("\n".join(metadata_lines))
        raw_data = read_csv(csv_stream, header=None, sep="=", names=["index", "values"])
        metadata = pd.Series(raw_data["values"].values, index=raw_data["index"])
        metadata.index = metadata.index.str.strip()

        reader.drop_empty()

        return result, ResultMetadata.create(
            SeriesData(metadata.str.strip()), experiment_type
        )

    @staticmethod
    def create_results(
        data: pd.DataFrame, experiment_type: ExperimentType
    ) -> dict[int, dict[str, Result]]:
        target_key = "Target Name"
        if experiment_type == ExperimentType.genotyping_qPCR_experiment:
            target_key = "SNP Assay Name"

        def make_results(well_data: pd.DataFrame) -> dict[str, Result]:
            return {
                target_dna_description: result
                for target_id, target_data in well_data.groupby(target_key)
                for target_dna_description, result in Result.create_result(
                    df_to_series_data(
                        target_data, msg="Unable to find parser result data"
                    ),
                    experiment_type,
                    str(target_id),
                ).items()
            }

        return map_wells(make_results, data)

    @staticmethod
    def create_result(
        data: SeriesData, experiment_type: ExperimentType, target_id: str
    ) -> dict[str, Result]:
        if experiment_type == ExperimentType.genotyping_qPCR_experiment:
            allele_prefixes = []
            for column in data.series.index:
                if match := re.match("(^\\w+) Ct$", column):
                    allele_prefixes.append(f"{match.groups()[0]} ")
        else:
            allele_prefixes = [""]

        return {
            f"{target_id}{f'-{allele_prefix}' if allele_prefix else ''}".replace(
                " ", ""
            ): Result(
                cycle_threshold_value_setting=data[
                    float, f"{allele_prefix}Ct Threshold"
                ],
                # TODO(nstender): really seems like this should be NaN if invalid value. Keeping to preserve tests.
                cycle_threshold_result=data.get(
                    float, [f"{allele_prefix}Ct", f"{allele_prefix}CT"]
                ),
                automatic_cycle_threshold_enabled_setting=data.get(
                    bool, f"{allele_prefix}Automatic Ct Threshold"
                ),
                automatic_baseline_determination_enabled_setting=data.get(
                    bool, f"{allele_prefix}Automatic Baseline"
                ),
                normalized_reporter_result=data.get(float, "Rn"),
                baseline_corrected_reporter_result=data.get(
                    float, f"{allele_prefix}Delta Rn"
                ),
                genotyping_determination_result=data.get(str, "Call"),
                genotyping_determination_method_setting=data.get(
                    float, "Threshold Value"
                ),
                quantity=data.get(float, "Quantity"),
                quantity_mean=data.get(float, "Quantity Mean"),
                quantity_sd=data.get(float, "Quantity SD"),
                ct_mean=data.get(float, "Ct Mean"),
                ct_sd=data.get(float, "Ct SD"),
                delta_ct_mean=data.get(float, "Delta Ct Mean"),
                delta_ct_se=data.get(float, "Delta Ct SE"),
                delta_delta_ct=data.get(float, "Delta Delta Ct"),
                rq=data.get(float, "RQ"),
                rq_min=data.get(float, "RQ Min"),
                rq_max=data.get(float, "RQ Max"),
                rn_mean=data.get(float, "Rn Mean"),
                rn_sd=data.get(float, "Rn SD"),
                y_intercept=data.get(float, "Y-Intercept"),
                r_squared=data.get(float, "R(superscript 2)"),
                slope=data.get(float, "Slope"),
                efficiency=data.get(float, "Efficiency"),
            )
            for allele_prefix in allele_prefixes
        }


@dataclass(frozen=True)
class MeltCurveRawData:
    reading: list[float]
    fluorescence: list[float | None]
    derivative: list[float | None]

    @staticmethod
    def create(reader: LinesReader) -> dict[int, MeltCurveRawData]:
        if not reader.match(r"^\[Melt Curve Raw Data\]"):
            return {}
        reader.pop()  # remove title
        lines = list(reader.pop_until_empty())
        csv_stream = StringIO("\n".join(lines))
        data = read_csv(csv_stream, sep="\t", thousands=r",")

        def make_data(well_data: pd.Series[Any]) -> MeltCurveRawData:
            return MeltCurveRawData(
                reading=well_data["Reading"].tolist(),
                fluorescence=well_data["Fluorescence"].tolist(),
                derivative=well_data["Derivative"].tolist(),
            )

        return map_wells(make_data, data)



from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.mabtech_apex.mabtech_apex_contents import MabtechApexContents
from allotropy.parsers.mabtech_apex.mabtech_apex_structure import create_data
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class MabtechApexParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "Mabtech Apex"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        contents = MabtechApexContents.create(named_file_contents)
        data = create_data(contents, named_file_contents.original_file_name)
        return self._get_mapper(Mapper).map_model(data)



from __future__ import annotations

from dataclasses import dataclass

import pandas as pd

from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    CalculatedDataItem,
    Data,
    DataSource,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ProcessedData,
    ProcessedDataFeature,
)
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.utils.iterables import get_first_not_none
from allotropy.parsers.utils.pandas import map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str

# These may be reported in the results, and are stored as calculated data
ABSORBANCE_RATIOS = [
    (260, 280),
    (260, 230),
]


@dataclass
class SpectroscopyRow:
    analyst: str | None
    timestamp: str
    experiment_type: str | None
    measurements: list[Measurement]
    calculated_data: list[CalculatedDataItem]

    @staticmethod
    def create(data: SeriesData) -> SpectroscopyRow:
        analyst = data.get(str, "user id")
        timestamp = f'{data[str, "date"]} {data.get(str, "time")}'
        experiment_type = data.get(str, "na type")

        sample_id = data.get(str, "sample id", NOT_APPLICABLE, SeriesData.NOT_NAN)
        well_plate_id = data.get(str, "plate id", validate=SeriesData.NOT_NAN)
        location_id = data[str, "well"]

        is_na_experiment = experiment_type and "NA" in experiment_type

        a260_absorbance = data.get(float, "a260")
        a280_absorbance = get_first_not_none(
            lambda key: data.get(float, key), ["a280", "a280 10mm"]
        )
        measurements: list[Measurement] = []

        mass_concentration = get_first_not_none(
            lambda key: data.get(float, key),
            ["conc.", "conc", "concentration"],
        )
        unit = data.get(str, "units")

        # We only capture mass concentration on one measurement document
        # TODO(nstender): why not just capture in both? Seems relevant, and would make this so much simpler.
        # capture concentration on the 260 measurement document if:
        #   - there is no experiment type and no 280 column add the concentration here
        # capture concentration on the 280 measurement document if:
        #   - the experiment type is something other than DNA or RNA
        #   - if the experiment type is not specified
        absorbances = (
            (
                260,
                a260_absorbance,
                is_na_experiment
                or not (experiment_type or a280_absorbance is not None),
            ),
            (280, a280_absorbance, not (experiment_type and is_na_experiment)),
        )

        for wavelength, absorbance, capture_concentration in absorbances:
            if not absorbance:
                continue
            measurements.append(
                Measurement(
                    type_=MeasurementType.ULTRAVIOLET_ABSORBANCE,
                    identifier=random_uuid_str(),
                    absorbance=absorbance,
                    detector_wavelength_setting=wavelength,
                    sample_identifier=sample_id,
                    well_plate_identifier=well_plate_id,
                    location_identifier=location_id,
                    processed_data=ProcessedData(
                        features=[
                            ProcessedDataFeature(
                                result=mass_concentration,
                                unit=unit,
                            )
                        ],
                    )
                    if capture_concentration and mass_concentration and unit
                    else None,
                )
            )

        absorbance_ratios = {}
        for numerator, denominator in ABSORBANCE_RATIOS:
            ratio = data.get(float, f"{numerator}/{denominator}")
            if ratio:
                absorbance_ratios[(numerator, denominator)] = ratio

        calculated_data = [
            CalculatedDataItem(
                identifier=random_uuid_str(),
                name=f"A{numerator}/{denominator}",
                value=ratio,
                unit=UNITLESS,
                data_sources=[
                    DataSource(identifier=measurement.identifier, feature="absorbance")
                    for measurement in measurements
                    if measurement.detector_wavelength_setting
                    in (numerator, denominator)
                ],
            )
            for (numerator, denominator), ratio in absorbance_ratios.items()
        ]

        return SpectroscopyRow(
            analyst,
            timestamp,
            experiment_type,
            measurements,
            calculated_data,
        )

    @staticmethod
    def create_rows(data: pd.DataFrame) -> list[SpectroscopyRow]:
        data.columns = data.columns.str.lower()
        return map_rows(data, SpectroscopyRow.create)


def create_data(data: pd.DataFrame, file_name: str) -> Data:
    rows = SpectroscopyRow.create_rows(data)

    return Data(
        metadata=Metadata(
            device_identifier="Nanodrop",
            device_type="absorbance detector",
            model_number="Nanodrop Eight",
            file_name=file_name,
        ),
        measurement_groups=[
            MeasurementGroup(
                measurement_time=row.timestamp,
                analyst=row.analyst,
                experiment_type=row.experiment_type,
                measurements=row.measurements,
            )
            for row in rows
        ],
        # NOTE: in current implementation, calculated data is reported at global level for some reason.
        # TODO(nstender): should we move this inside of measurements?
        calculated_data=[item for row in rows for item in row.calculated_data],
    )



from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import ExperimentType
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_calculated_documents import (
    iter_presence_absence_calc_docs,
    iter_relative_standard_curve_calc_docs,
    iter_standard_curve_calc_docs,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_contents import (
    DesignQuantstudioContents,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_structure import (
    Data,
    Header,
    Result,
    WellList,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_views import (
    SampleView,
    TargetRoleView,
    TargetView,
)


def create_data(contents: DesignQuantstudioContents) -> Data:
    experiment_type = Data.get_experiment_type(contents)
    header = Header.create(contents.header)
    wells = WellList.create(contents, header, experiment_type)
    well_items = wells.get_well_items()

    view_st_data = SampleView(sub_view=TargetView()).apply(well_items)
    r_sample = None
    r_target = None

    if experiment_type == ExperimentType.standard_curve_qPCR_experiment:
        calculated_documents = list(
            iter_standard_curve_calc_docs(
                view_st_data,
                view_tr_data=TargetRoleView().apply(well_items),
            )
        )
    elif experiment_type == ExperimentType.relative_standard_curve_qPCR_experiment:
        r_sample = Result.get_reference_sample(contents)
        r_target = Result.get_reference_target(contents)
        calculated_documents = list(
            iter_relative_standard_curve_calc_docs(
                view_st_data,
                r_sample,
                r_target,
            )
        )
    elif experiment_type == ExperimentType.presence_absence_qPCR_experiment:
        calculated_documents = list(
            iter_presence_absence_calc_docs(
                view_st_data,
            )
        )
    else:
        calculated_documents = []

    return Data(
        header,
        wells,
        experiment_type,
        calculated_documents,
        reference_target=r_target,
        reference_sample=r_sample,
    )



from __future__ import annotations

from dataclasses import dataclass

import pandas as pd

from allotropy.parsers.lines_reader import CsvReader
from allotropy.parsers.utils.values import assert_not_none

FILENAME = "FileName"
BARCODE_1 = "Barcode1"
READ_TIME = "Read Time"
VERSION = "Version"
USER = "User"
SERIAL_NO = "Serial"
MODEL = "Model"
DATA = "Data"
PLATE_END = "==="


@dataclass(frozen=True)
class CombinedData:
    file_name: str
    version: str
    model: str
    serial_number: str
    plate_doc_info: list[PlateData]

    @staticmethod
    def create(reader: CsvReader) -> CombinedData:
        file_name = assert_not_none(CombinedData.get_parameter(reader, FILENAME))
        barcode_1 = assert_not_none(
            CombinedData.get_parameter(reader, BARCODE_1)
        ).strip("<>")
        read_time = assert_not_none(
            CombinedData.get_parameter(reader, READ_TIME, all_vals_after_tab=True)
        )
        version = assert_not_none(CombinedData.get_parameter(reader, VERSION))
        user = CombinedData.get_parameter(reader, USER)
        serial_no = assert_not_none(CombinedData.get_parameter(reader, SERIAL_NO))
        model = assert_not_none(CombinedData.get_parameter(reader, MODEL))

        plate_data = []
        while reader.current_line < len(reader.lines):
            assert_not_none(
                reader.drop_until_inclusive(DATA),
                "Data",
                "Missing Data section in file",
            )
            plate_start = reader.current_line
            reader.drop_until(PLATE_END)
            plate_end = reader.current_line

            # Create dataframe of the plate, drop any columns with all Nan values
            plate_df = assert_not_none(
                reader.lines_as_df(
                    reader.lines[plate_start:plate_end], sep="\t", header=0
                ),
                "Luminescence data table",
            ).dropna(axis=1, how="all")

            # Drop rows with nans in the first column since they don't represent a row in a plate
            # Drop the first column to get total wells since it's just a "label"
            dropped_nans = plate_df.dropna(subset=[plate_df.columns[0]]).drop(
                columns=[plate_df.columns[0]]
            )
            # Get the well count by multiplying number of rows by number of cols
            well_count = dropped_nans.shape[0] * dropped_nans.shape[1]
            plate_data.append(
                PlateData.create(
                    plate_df=plate_df,
                    measurement_time=read_time,
                    analyst=user,
                    well_plate_id=barcode_1,
                    plate_well_count=well_count,
                )
            )

            # Attempt to get the next plate's barcode, read time, and analyst.
            next_barcode = reader.drop_until_inclusive(BARCODE_1)
            next_read_time = reader.drop_until_inclusive(READ_TIME)
            next_user = reader.drop_until_inclusive(USER)

            if all(
                var is not None for var in [next_barcode, next_read_time, next_user]
            ):
                barcode_1 = assert_not_none(next_barcode).split("\t")[1].strip("<>")
                read_time = " ".join(assert_not_none(next_read_time).split("\t")[1:])
                user = assert_not_none(next_user).split("\t")[1]
            # Exception if the next plate isn't found, stop looking for more plates
            else:
                break
        return CombinedData(
            file_name=file_name,
            version=version,
            model=model,
            serial_number=serial_no,
            plate_doc_info=plate_data,
        )

    @staticmethod
    def get_parameter(
        reader: CsvReader, name: str, *, all_vals_after_tab: bool = False
    ) -> str | None:
        val_line = reader.drop_until_inclusive(name)
        if val_line is not None:
            try:
                if all_vals_after_tab:
                    return " ".join(val_line.split("\t")[1:])
                else:
                    return val_line.split("\t")[1]
            except IndexError:
                return None
        else:
            return None


@dataclass(frozen=True)
class PlateData:
    measurement_time: str
    plate_well_count: int
    analyst: str | None
    well_plate_id: str
    well_data: list[WellData]

    @staticmethod
    def create(
        plate_df: pd.DataFrame,
        measurement_time: str,
        analyst: str | None,
        well_plate_id: str,
        plate_well_count: int,
    ) -> PlateData:
        well_data = []
        spot_index_counter = 0
        for _index, row in plate_df.iterrows():
            # Increment the spot index counter for every new row
            spot_index_counter += 1
            # Skip the first column since it's the well names (no luminescence values)
            for col in plate_df.columns[1:]:
                if pd.notna(row.iloc[0]):
                    # This is the row well name-- A, B, C, etc.
                    well_row = row.iloc[0].strip()
                    # If we've detected a new well row, reset the spot index counter
                    spot_index_counter = 1
                location_name = well_row + col.strip() + "_" + str(spot_index_counter)
                luminescence = row[col]
                well_data.append(
                    WellData.create(
                        luminescence=luminescence,
                        location_id=location_name,
                        well_plate_id=well_plate_id,
                    )
                )
        return PlateData(
            measurement_time=measurement_time,
            plate_well_count=plate_well_count,
            analyst=analyst,
            well_plate_id=well_plate_id,
            well_data=well_data,
        )


@dataclass(frozen=True)
class WellData:
    luminescence: int
    location_identifier: str
    sample_identifier: str

    @staticmethod
    def create(luminescence: int, location_id: str, well_plate_id: str) -> WellData:
        sample_id = well_plate_id + "_" + location_id
        return WellData(
            luminescence=luminescence,
            location_identifier=location_id,
            sample_identifier=sample_id,
        )



from __future__ import annotations

from typing import Any

import pandas as pd
from pandas._typing import FilePath, ReadCsvBuffer

from allotropy.exceptions import AllotropeConversionError, AllotropeParsingError
from allotropy.types import IOType


def read_csv(
    # types for filepath_or_buffer match those in pd.read_csv()
    filepath_or_buffer: FilePath | ReadCsvBuffer[bytes] | ReadCsvBuffer[str],
    **kwargs: Any,
) -> pd.DataFrame:
    """Wrap pd.read_csv() and raise AllotropeParsingError for failures.

    pd.read_csv() can return a DataFrame or TextFileReader. The latter is intentionally not supported."""
    try:
        df_or_reader = pd.read_csv(filepath_or_buffer, **kwargs)
    except Exception as e:
        msg = f"Error calling pd.read_csv(): {e}"
        raise AllotropeParsingError(msg) from e
    return _df_or_bust(
        df_or_reader, "pd.read_csv() returned a TextFileReader, which is not supported."
    )


def read_excel(
    # io is untyped in pd.read_excel(), but this seems reasonable.
    io: str | IOType,
    **kwargs: Any,
) -> pd.DataFrame:
    """Wrap pd.read_excel() and raise AllotropeParsingError for failures.

    pd.read_excel() can return a DataFrame or a dictionary of DataFrames. The latter is intentionally not supported."""
    try:
        df_or_dict = pd.read_excel(io, **kwargs)
    except Exception as e:
        msg = f"Error calling pd.read_excel(): {e}"
        raise AllotropeParsingError(msg) from e
    return _df_or_bust(df_or_dict, "Expected a single-sheet Excel file.")


def _df_or_bust(possible_df: Any, error_message: str) -> pd.DataFrame:
    if isinstance(possible_df, pd.DataFrame):
        return possible_df
    raise AllotropeConversionError(error_message)



from __future__ import annotations

from allotropy.allotrope.models.adm.cell_counting.benchling._2023._11.cell_counting import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.cell_counting.benchling._2023._11.cell_counting import (
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.beckman_vi_cell_blu.vi_cell_blu_reader import ViCellBluReader
from allotropy.parsers.beckman_vi_cell_blu.vi_cell_blu_structure import create_data
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class ViCellBluParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "Beckman Vi-Cell BLU"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        data = create_data(
            ViCellBluReader.read(named_file_contents),
            named_file_contents.original_file_name,
        )
        return self._get_mapper(Mapper).map_model(data)



from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.ctl_immunospot.ctl_immunospot_structure import create_data
from allotropy.parsers.lines_reader import LinesReader, read_to_lines
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class CtlImmunospotParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "CTL ImmunoSpot"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        lines = read_to_lines(named_file_contents)
        data = create_data(LinesReader(lines))
        return self._get_mapper(Mapper).map_model(data)



NUCLEOCOUNTER_SOFTWARE_NAME = "NucleoView"
NUCLEOCOUNTER_DEVICE_TYPE = "dark field imager (cell counter)"
NUCLEOCOUNTER_DETECTION_TYPE = "dark field"

DEFAULT_ANALYST = "NC-200"
DEFAULT_MODEL_NUMBER = "NucleoCounter NC-200"
DEFAULT_EPOCH_TIMESTAMP = "1970-01-01"



#!/usr/bin/env python3
import click

from allotropy.parser_factory import update_readme


@click.command()
def _update_readme() -> None:
    """Update parser list in README."""
    update_readme()


if __name__ == "__main__":
    _update_readme()



# mypy: disallow_any_generics = False

from collections.abc import Callable
from typing import Any


def cache(fun: Callable) -> Callable:
    cache: dict[str, Any] = {}

    def inner(*args: Any) -> Any:
        key = "".join([str(arg) for arg in args])
        if key in cache:
            return cache[key]
        result = fun(*args)
        cache[key] = result
        return result

    return inner
